# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   Partial C706.pdf + [MS-RPCE] implementation
#
#   Best way to learn how to use these calls is to grab the protocol standard
#   so you understand what the call does, and then read the test case located
#   at https://github.com/SecureAuthCorp/impacket/tree/master/tests/SMB_RPC
#
# ToDo: 
# [ ] Take out all the security provider stuff out of here (e.g. RPC_C_AUTHN_WINNT)
#	 and put it elsewhere. This will make the coder cleaner and easier to add 
#	 more SSP (e.g. NETLOGON)
# 

import asyncio
import logging
import socket
import sys
from binascii import unhexlify

from aiosmb.crypto.symmetric import RC4

from aiosmb import logger as LOG


from aiosmb.dcerpc.v5.structure import Structure,pack,unpack
from aiosmb.dcerpc.v5 import uuid
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin, generate, stringver_to_bin, bin_to_uuidtup
from aiosmb.dcerpc.v5.dtypes import UCHAR, ULONG, USHORT
from aiosmb.dcerpc.v5.ndr import NDRSTRUCT
from aiosmb.dcerpc.v5 import hresult_errors
from minikerberos.gssapi import *


# MS/RPC Constants
MSRPC_REQUEST   = 0x00
MSRPC_PING	  = 0x01
MSRPC_RESPONSE  = 0x02
MSRPC_FAULT	 = 0x03
MSRPC_WORKING   = 0x04
MSRPC_NOCALL	= 0x05
MSRPC_REJECT	= 0x06
MSRPC_ACK	   = 0x07
MSRPC_CL_CANCEL = 0x08
MSRPC_FACK	  = 0x09
MSRPC_CANCELACK = 0x0A
MSRPC_BIND	  = 0x0B
MSRPC_BINDACK   = 0x0C
MSRPC_BINDNAK   = 0x0D
MSRPC_ALTERCTX  = 0x0E
MSRPC_ALTERCTX_R= 0x0F
MSRPC_AUTH3	 = 0x10
MSRPC_SHUTDOWN  = 0x11
MSRPC_CO_CANCEL = 0x12
MSRPC_ORPHANED  = 0x13

# MS/RPC Packet Flags
PFC_FIRST_FRAG	 = 0x01
PFC_LAST_FRAG	  = 0x02

# For PDU types bind, bind_ack, alter_context, and
# alter_context_resp, this flag MUST be interpreted as PFC_SUPPORT_HEADER_SIGN
MSRPC_SUPPORT_SIGN  = 0x04

#For the
#remaining PDU types, this flag MUST be interpreted as PFC_PENDING_CANCEL.
MSRPC_PENDING_CANCEL= 0x04

PFC_RESERVED_1	  = 0x08
PFC_CONC_MPX		= 0x10
PFC_DID_NOT_EXECUTE = 0x20
PFC_MAYBE		   = 0x40
PFC_OBJECT_UUID	 = 0x80

# Auth Types - Security Providers
RPC_C_AUTHN_NONE		  = 0x00
RPC_C_AUTHN_GSS_NEGOTIATE = 0x09
RPC_C_AUTHN_WINNT		 = 0x0A
RPC_C_AUTHN_GSS_SCHANNEL  = 0x0E
RPC_C_AUTHN_GSS_KERBEROS  = 0x10
RPC_C_AUTHN_NETLOGON	  = 0x44
RPC_C_AUTHN_DEFAULT	   = 0xFF

# Auth Levels
RPC_C_AUTHN_LEVEL_NONE		  = 1
RPC_C_AUTHN_LEVEL_CONNECT	   = 2
RPC_C_AUTHN_LEVEL_CALL		  = 3
RPC_C_AUTHN_LEVEL_PKT		   = 4
RPC_C_AUTHN_LEVEL_PKT_INTEGRITY = 5
RPC_C_AUTHN_LEVEL_PKT_PRIVACY   = 6

#Reasons for rejection of a context element, included in bind_ack result reason
rpc_provider_reason = {
	0	   : 'reason_not_specified',
	1	   : 'abstract_syntax_not_supported',
	2	   : 'proposed_transfer_syntaxes_not_supported',
	3	   : 'local_limit_exceeded',
	4	   : 'protocol_version_not_specified',
	8	   : 'authentication_type_not_recognized',
	9	   : 'invalid_checksum'
}

MSRPC_CONT_RESULT_ACCEPT = 0
MSRPC_CONT_RESULT_USER_REJECT = 1
MSRPC_CONT_RESULT_PROV_REJECT = 2

#Results of a presentation context negotiation
rpc_cont_def_result = {
	0	   : 'acceptance',
	1	   : 'user_rejection',
	2	   : 'provider_rejection'
}

#status codes, references:
#https://docs.microsoft.com/windows/desktop/Rpc/rpc-return-values
#https://msdn.microsoft.com/library/default.asp?url=/library/en-us/randz/protocol/common_return_values.asp
#winerror.h
#https://www.opengroup.org/onlinepubs/9629399/apdxn.htm

rpc_status_codes = {
	0x00000005 : 'rpc_s_access_denied',
	0x00000008 : 'Authentication type not recognized',
	0x000006D8 : 'rpc_fault_cant_perform', 
	0x000006C6 : 'rpc_x_invalid_bound',				# the arrays bound are invalid
	0x000006E4 : 'rpc_s_cannot_support: The requested operation is not supported.',			   # some operation is not supported
	0x000006F7 : 'rpc_x_bad_stub_data',				# the stub data is invalid, doesn't match with the IDL definition
	0x1C010001 : 'nca_s_comm_failure',				 # unable to get response from server:
	0x1C010002 : 'nca_s_op_rng_error',				 # bad operation number in call
	0x1C010003 : 'nca_s_unk_if',					   # unknown interface
	0x1C010006 : 'nca_s_wrong_boot_time',			  # client passed server wrong server boot time
	0x1C010009 : 'nca_s_you_crashed',				  # a restarted server called back a client
	0x1C01000B : 'nca_s_proto_error',				  # someone messed up the protocol
	0x1C010013 : 'nca_s_out_args_too_big ',			# output args too big
	0x1C010014 : 'nca_s_server_too_busy',			  # server is too busy to handle call
	0x1C010015 : 'nca_s_fault_string_too_long',		# string argument longer than declared max len
	0x1C010017 : 'nca_s_unsupported_type ',			# no implementation of generic operation for object
	0x1C000001 : 'nca_s_fault_int_div_by_zero',
	0x1C000002 : 'nca_s_fault_addr_error ',
	0x1C000003 : 'nca_s_fault_fp_div_zero',
	0x1C000004 : 'nca_s_fault_fp_underflow',
	0x1C000005 : 'nca_s_fault_fp_overflow',
	0x1C000006 : 'nca_s_fault_invalid_tag',
	0x1C000007 : 'nca_s_fault_invalid_bound ',
	0x1C000008 : 'nca_s_rpc_version_mismatch',
	0x1C000009 : 'nca_s_unspec_reject ',
	0x1C00000A : 'nca_s_bad_actid',
	0x1C00000B : 'nca_s_who_are_you_failed',
	0x1C00000C : 'nca_s_manager_not_entered ',
	0x1C00000D : 'nca_s_fault_cancel',
	0x1C00000E : 'nca_s_fault_ill_inst',
	0x1C00000F : 'nca_s_fault_fp_error',
	0x1C000010 : 'nca_s_fault_int_overflow',
	0x1C000012 : 'nca_s_fault_unspec',
	0x1C000013 : 'nca_s_fault_remote_comm_failure ',
	0x1C000014 : 'nca_s_fault_pipe_empty ',
	0x1C000015 : 'nca_s_fault_pipe_closed',
	0x1C000016 : 'nca_s_fault_pipe_order ',
	0x1C000017 : 'nca_s_fault_pipe_discipline',
	0x1C000018 : 'nca_s_fault_pipe_comm_error',
	0x1C000019 : 'nca_s_fault_pipe_memory',
	0x1C00001A : 'nca_s_fault_context_mismatch ',
	0x1C00001B : 'nca_s_fault_remote_no_memory ',
	0x1C00001C : 'nca_s_invalid_pres_context_id',
	0x1C00001D : 'nca_s_unsupported_authn_level',
	0x1C00001F : 'nca_s_invalid_checksum ',
	0x1C000020 : 'nca_s_invalid_crc',
	0x1C000021 : 'nca_s_fault_user_defined',
	0x1C000022 : 'nca_s_fault_tx_open_failed',
	0x1C000023 : 'nca_s_fault_codeset_conv_error',
	0x1C000024 : 'nca_s_fault_object_not_found ',
	0x1C000025 : 'nca_s_fault_no_client_stub',
	0x16c9a000 : "rpc_s_mod",
	0x16c9a001 : "rpc_s_op_rng_error",
	0x16c9a002 : "rpc_s_cant_create_socket",
	0x16c9a003 : "rpc_s_cant_bind_socket",
	0x16c9a004 : "rpc_s_not_in_call",
	0x16c9a005 : "rpc_s_no_port",
	0x16c9a006 : "rpc_s_wrong_boot_time",
	0x16c9a007 : "rpc_s_too_many_sockets",
	0x16c9a008 : "rpc_s_illegal_register",
	0x16c9a009 : "rpc_s_cant_recv",
	0x16c9a00a : "rpc_s_bad_pkt",
	0x16c9a00b : "rpc_s_unbound_handle",
	0x16c9a00c : "rpc_s_addr_in_use",
	0x16c9a00d : "rpc_s_in_args_too_big",
	0x16c9a00e : "rpc_s_string_too_long",
	0x16c9a00f : "rpc_s_too_many_objects",
	0x16c9a010 : "rpc_s_binding_has_no_auth",
	0x16c9a011 : "rpc_s_unknown_authn_service",
	0x16c9a012 : "rpc_s_no_memory",
	0x16c9a013 : "rpc_s_cant_nmalloc",
	0x16c9a014 : "rpc_s_call_faulted",
	0x16c9a015 : "rpc_s_call_failed",
	0x16c9a016 : "rpc_s_comm_failure",
	0x16c9a017 : "rpc_s_rpcd_comm_failure",
	0x16c9a018 : "rpc_s_illegal_family_rebind",
	0x16c9a019 : "rpc_s_invalid_handle",
	0x16c9a01a : "rpc_s_coding_error",
	0x16c9a01b : "rpc_s_object_not_found",
	0x16c9a01c : "rpc_s_cthread_not_found",
	0x16c9a01d : "rpc_s_invalid_binding",
	0x16c9a01e : "rpc_s_already_registered",
	0x16c9a01f : "rpc_s_endpoint_not_found",
	0x16c9a020 : "rpc_s_invalid_rpc_protseq",
	0x16c9a021 : "rpc_s_desc_not_registered",
	0x16c9a022 : "rpc_s_already_listening",
	0x16c9a023 : "rpc_s_no_protseqs",
	0x16c9a024 : "rpc_s_no_protseqs_registered",
	0x16c9a025 : "rpc_s_no_bindings",
	0x16c9a026 : "rpc_s_max_descs_exceeded",
	0x16c9a027 : "rpc_s_no_interfaces",
	0x16c9a028 : "rpc_s_invalid_timeout",
	0x16c9a029 : "rpc_s_cant_inq_socket",
	0x16c9a02a : "rpc_s_invalid_naf_id",
	0x16c9a02b : "rpc_s_inval_net_addr",
	0x16c9a02c : "rpc_s_unknown_if",
	0x16c9a02d : "rpc_s_unsupported_type",
	0x16c9a02e : "rpc_s_invalid_call_opt",
	0x16c9a02f : "rpc_s_no_fault",
	0x16c9a030 : "rpc_s_cancel_timeout",
	0x16c9a031 : "rpc_s_call_cancelled",
	0x16c9a032 : "rpc_s_invalid_call_handle",
	0x16c9a033 : "rpc_s_cannot_alloc_assoc",
	0x16c9a034 : "rpc_s_cannot_connect",
	0x16c9a035 : "rpc_s_connection_aborted",
	0x16c9a036 : "rpc_s_connection_closed",
	0x16c9a037 : "rpc_s_cannot_accept",
	0x16c9a038 : "rpc_s_assoc_grp_not_found",
	0x16c9a039 : "rpc_s_stub_interface_error",
	0x16c9a03a : "rpc_s_invalid_object",
	0x16c9a03b : "rpc_s_invalid_type",
	0x16c9a03c : "rpc_s_invalid_if_opnum",
	0x16c9a03d : "rpc_s_different_server_instance",
	0x16c9a03e : "rpc_s_protocol_error",
	0x16c9a03f : "rpc_s_cant_recvmsg",
	0x16c9a040 : "rpc_s_invalid_string_binding",
	0x16c9a041 : "rpc_s_connect_timed_out",
	0x16c9a042 : "rpc_s_connect_rejected",
	0x16c9a043 : "rpc_s_network_unreachable",
	0x16c9a044 : "rpc_s_connect_no_resources",
	0x16c9a045 : "rpc_s_rem_network_shutdown",
	0x16c9a046 : "rpc_s_too_many_rem_connects",
	0x16c9a047 : "rpc_s_no_rem_endpoint",
	0x16c9a048 : "rpc_s_rem_host_down",
	0x16c9a049 : "rpc_s_host_unreachable",
	0x16c9a04a : "rpc_s_access_control_info_inv",
	0x16c9a04b : "rpc_s_loc_connect_aborted",
	0x16c9a04c : "rpc_s_connect_closed_by_rem",
	0x16c9a04d : "rpc_s_rem_host_crashed",
	0x16c9a04e : "rpc_s_invalid_endpoint_format",
	0x16c9a04f : "rpc_s_unknown_status_code",
	0x16c9a050 : "rpc_s_unknown_mgr_type",
	0x16c9a051 : "rpc_s_assoc_creation_failed",
	0x16c9a052 : "rpc_s_assoc_grp_max_exceeded",
	0x16c9a053 : "rpc_s_assoc_grp_alloc_failed",
	0x16c9a054 : "rpc_s_sm_invalid_state",
	0x16c9a055 : "rpc_s_assoc_req_rejected",
	0x16c9a056 : "rpc_s_assoc_shutdown",
	0x16c9a057 : "rpc_s_tsyntaxes_unsupported",
	0x16c9a058 : "rpc_s_context_id_not_found",
	0x16c9a059 : "rpc_s_cant_listen_socket",
	0x16c9a05a : "rpc_s_no_addrs",
	0x16c9a05b : "rpc_s_cant_getpeername",
	0x16c9a05c : "rpc_s_cant_get_if_id",
	0x16c9a05d : "rpc_s_protseq_not_supported",
	0x16c9a05e : "rpc_s_call_orphaned",
	0x16c9a05f : "rpc_s_who_are_you_failed",
	0x16c9a060 : "rpc_s_unknown_reject",
	0x16c9a061 : "rpc_s_type_already_registered",
	0x16c9a062 : "rpc_s_stop_listening_disabled",
	0x16c9a063 : "rpc_s_invalid_arg",
	0x16c9a064 : "rpc_s_not_supported",
	0x16c9a065 : "rpc_s_wrong_kind_of_binding",
	0x16c9a066 : "rpc_s_authn_authz_mismatch",
	0x16c9a067 : "rpc_s_call_queued",
	0x16c9a068 : "rpc_s_cannot_set_nodelay",
	0x16c9a069 : "rpc_s_not_rpc_tower",
	0x16c9a06a : "rpc_s_invalid_rpc_protid",
	0x16c9a06b : "rpc_s_invalid_rpc_floor",
	0x16c9a06c : "rpc_s_call_timeout",
	0x16c9a06d : "rpc_s_mgmt_op_disallowed",
	0x16c9a06e : "rpc_s_manager_not_entered",
	0x16c9a06f : "rpc_s_calls_too_large_for_wk_ep",
	0x16c9a070 : "rpc_s_server_too_busy",
	0x16c9a071 : "rpc_s_prot_version_mismatch",
	0x16c9a072 : "rpc_s_rpc_prot_version_mismatch",
	0x16c9a073 : "rpc_s_ss_no_import_cursor",
	0x16c9a074 : "rpc_s_fault_addr_error",
	0x16c9a075 : "rpc_s_fault_context_mismatch",
	0x16c9a076 : "rpc_s_fault_fp_div_by_zero",
	0x16c9a077 : "rpc_s_fault_fp_error",
	0x16c9a078 : "rpc_s_fault_fp_overflow",
	0x16c9a079 : "rpc_s_fault_fp_underflow",
	0x16c9a07a : "rpc_s_fault_ill_inst",
	0x16c9a07b : "rpc_s_fault_int_div_by_zero",
	0x16c9a07c : "rpc_s_fault_int_overflow",
	0x16c9a07d : "rpc_s_fault_invalid_bound",
	0x16c9a07e : "rpc_s_fault_invalid_tag",
	0x16c9a07f : "rpc_s_fault_pipe_closed",
	0x16c9a080 : "rpc_s_fault_pipe_comm_error",
	0x16c9a081 : "rpc_s_fault_pipe_discipline",
	0x16c9a082 : "rpc_s_fault_pipe_empty",
	0x16c9a083 : "rpc_s_fault_pipe_memory",
	0x16c9a084 : "rpc_s_fault_pipe_order",
	0x16c9a085 : "rpc_s_fault_remote_comm_failure",
	0x16c9a086 : "rpc_s_fault_remote_no_memory",
	0x16c9a087 : "rpc_s_fault_unspec",
	0x16c9a088 : "uuid_s_bad_version",
	0x16c9a089 : "uuid_s_socket_failure",
	0x16c9a08a : "uuid_s_getconf_failure",
	0x16c9a08b : "uuid_s_no_address",
	0x16c9a08c : "uuid_s_overrun",
	0x16c9a08d : "uuid_s_internal_error",
	0x16c9a08e : "uuid_s_coding_error",
	0x16c9a08f : "uuid_s_invalid_string_uuid",
	0x16c9a090 : "uuid_s_no_memory",
	0x16c9a091 : "rpc_s_no_more_entries",
	0x16c9a092 : "rpc_s_unknown_ns_error",
	0x16c9a093 : "rpc_s_name_service_unavailable",
	0x16c9a094 : "rpc_s_incomplete_name",
	0x16c9a095 : "rpc_s_group_not_found",
	0x16c9a096 : "rpc_s_invalid_name_syntax",
	0x16c9a097 : "rpc_s_no_more_members",
	0x16c9a098 : "rpc_s_no_more_interfaces",
	0x16c9a099 : "rpc_s_invalid_name_service",
	0x16c9a09a : "rpc_s_no_name_mapping",
	0x16c9a09b : "rpc_s_profile_not_found",
	0x16c9a09c : "rpc_s_not_found",
	0x16c9a09d : "rpc_s_no_updates",
	0x16c9a09e : "rpc_s_update_failed",
	0x16c9a09f : "rpc_s_no_match_exported",
	0x16c9a0a0 : "rpc_s_entry_not_found",
	0x16c9a0a1 : "rpc_s_invalid_inquiry_context",
	0x16c9a0a2 : "rpc_s_interface_not_found",
	0x16c9a0a3 : "rpc_s_group_member_not_found",
	0x16c9a0a4 : "rpc_s_entry_already_exists",
	0x16c9a0a5 : "rpc_s_nsinit_failure",
	0x16c9a0a6 : "rpc_s_unsupported_name_syntax",
	0x16c9a0a7 : "rpc_s_no_more_elements",
	0x16c9a0a8 : "rpc_s_no_ns_permission",
	0x16c9a0a9 : "rpc_s_invalid_inquiry_type",
	0x16c9a0aa : "rpc_s_profile_element_not_found",
	0x16c9a0ab : "rpc_s_profile_element_replaced",
	0x16c9a0ac : "rpc_s_import_already_done",
	0x16c9a0ad : "rpc_s_database_busy",
	0x16c9a0ae : "rpc_s_invalid_import_context",
	0x16c9a0af : "rpc_s_uuid_set_not_found",
	0x16c9a0b0 : "rpc_s_uuid_member_not_found",
	0x16c9a0b1 : "rpc_s_no_interfaces_exported",
	0x16c9a0b2 : "rpc_s_tower_set_not_found",
	0x16c9a0b3 : "rpc_s_tower_member_not_found",
	0x16c9a0b4 : "rpc_s_obj_uuid_not_found",
	0x16c9a0b5 : "rpc_s_no_more_bindings",
	0x16c9a0b6 : "rpc_s_invalid_priority",
	0x16c9a0b7 : "rpc_s_not_rpc_entry",
	0x16c9a0b8 : "rpc_s_invalid_lookup_context",
	0x16c9a0b9 : "rpc_s_binding_vector_full",
	0x16c9a0ba : "rpc_s_cycle_detected",
	0x16c9a0bb : "rpc_s_nothing_to_export",
	0x16c9a0bc : "rpc_s_nothing_to_unexport",
	0x16c9a0bd : "rpc_s_invalid_vers_option",
	0x16c9a0be : "rpc_s_no_rpc_data",
	0x16c9a0bf : "rpc_s_mbr_picked",
	0x16c9a0c0 : "rpc_s_not_all_objs_unexported",
	0x16c9a0c1 : "rpc_s_no_entry_name",
	0x16c9a0c2 : "rpc_s_priority_group_done",
	0x16c9a0c3 : "rpc_s_partial_results",
	0x16c9a0c4 : "rpc_s_no_env_setup",
	0x16c9a0c5 : "twr_s_unknown_sa",
	0x16c9a0c6 : "twr_s_unknown_tower",
	0x16c9a0c7 : "twr_s_not_implemented",
	0x16c9a0c8 : "rpc_s_max_calls_too_small",
	0x16c9a0c9 : "rpc_s_cthread_create_failed",
	0x16c9a0ca : "rpc_s_cthread_pool_exists",
	0x16c9a0cb : "rpc_s_cthread_no_such_pool",
	0x16c9a0cc : "rpc_s_cthread_invoke_disabled",
	0x16c9a0cd : "ept_s_cant_perform_op",
	0x16c9a0ce : "ept_s_no_memory",
	0x16c9a0cf : "ept_s_database_invalid",
	0x16c9a0d0 : "ept_s_cant_create",
	0x16c9a0d1 : "ept_s_cant_access",
	0x16c9a0d2 : "ept_s_database_already_open",
	0x16c9a0d3 : "ept_s_invalid_entry",
	0x16c9a0d4 : "ept_s_update_failed",
	0x16c9a0d5 : "ept_s_invalid_context",
	0x16c9a0d6 : "ept_s_not_registered",
	0x16c9a0d7 : "ept_s_server_unavailable",
	0x16c9a0d8 : "rpc_s_underspecified_name",
	0x16c9a0d9 : "rpc_s_invalid_ns_handle",
	0x16c9a0da : "rpc_s_unknown_error",
	0x16c9a0db : "rpc_s_ss_char_trans_open_fail",
	0x16c9a0dc : "rpc_s_ss_char_trans_short_file",
	0x16c9a0dd : "rpc_s_ss_context_damaged",
	0x16c9a0de : "rpc_s_ss_in_null_context",
	0x16c9a0df : "rpc_s_socket_failure",
	0x16c9a0e0 : "rpc_s_unsupported_protect_level",
	0x16c9a0e1 : "rpc_s_invalid_checksum",
	0x16c9a0e2 : "rpc_s_invalid_credentials",
	0x16c9a0e3 : "rpc_s_credentials_too_large",
	0x16c9a0e4 : "rpc_s_call_id_not_found",
	0x16c9a0e5 : "rpc_s_key_id_not_found",
	0x16c9a0e6 : "rpc_s_auth_bad_integrity",
	0x16c9a0e7 : "rpc_s_auth_tkt_expired",
	0x16c9a0e8 : "rpc_s_auth_tkt_nyv",
	0x16c9a0e9 : "rpc_s_auth_repeat",
	0x16c9a0ea : "rpc_s_auth_not_us",
	0x16c9a0eb : "rpc_s_auth_badmatch",
	0x16c9a0ec : "rpc_s_auth_skew",
	0x16c9a0ed : "rpc_s_auth_badaddr",
	0x16c9a0ee : "rpc_s_auth_badversion",
	0x16c9a0ef : "rpc_s_auth_msg_type",
	0x16c9a0f0 : "rpc_s_auth_modified",
	0x16c9a0f1 : "rpc_s_auth_badorder",
	0x16c9a0f2 : "rpc_s_auth_badkeyver",
	0x16c9a0f3 : "rpc_s_auth_nokey",
	0x16c9a0f4 : "rpc_s_auth_mut_fail",
	0x16c9a0f5 : "rpc_s_auth_baddirection",
	0x16c9a0f6 : "rpc_s_auth_method",
	0x16c9a0f7 : "rpc_s_auth_badseq",
	0x16c9a0f8 : "rpc_s_auth_inapp_cksum",
	0x16c9a0f9 : "rpc_s_auth_field_toolong",
	0x16c9a0fa : "rpc_s_invalid_crc",
	0x16c9a0fb : "rpc_s_binding_incomplete",
	0x16c9a0fc : "rpc_s_key_func_not_allowed",
	0x16c9a0fd : "rpc_s_unknown_stub_rtl_if_vers",
	0x16c9a0fe : "rpc_s_unknown_ifspec_vers",
	0x16c9a0ff : "rpc_s_proto_unsupp_by_auth",
	0x16c9a100 : "rpc_s_authn_challenge_malformed",
	0x16c9a101 : "rpc_s_protect_level_mismatch",
	0x16c9a102 : "rpc_s_no_mepv",
	0x16c9a103 : "rpc_s_stub_protocol_error",
	0x16c9a104 : "rpc_s_class_version_mismatch",
	0x16c9a105 : "rpc_s_helper_not_running",
	0x16c9a106 : "rpc_s_helper_short_read",
	0x16c9a107 : "rpc_s_helper_catatonic",
	0x16c9a108 : "rpc_s_helper_aborted",
	0x16c9a109 : "rpc_s_not_in_kernel",
	0x16c9a10a : "rpc_s_helper_wrong_user",
	0x16c9a10b : "rpc_s_helper_overflow",
	0x16c9a10c : "rpc_s_dg_need_way_auth",
	0x16c9a10d : "rpc_s_unsupported_auth_subtype",
	0x16c9a10e : "rpc_s_wrong_pickle_type",
	0x16c9a10f : "rpc_s_not_listening",
	0x16c9a110 : "rpc_s_ss_bad_buffer",
	0x16c9a111 : "rpc_s_ss_bad_es_action",
	0x16c9a112 : "rpc_s_ss_wrong_es_version",
	0x16c9a113 : "rpc_s_fault_user_defined",
	0x16c9a114 : "rpc_s_ss_incompatible_codesets",
	0x16c9a115 : "rpc_s_tx_not_in_transaction",
	0x16c9a116 : "rpc_s_tx_open_failed",
	0x16c9a117 : "rpc_s_partial_credentials",
	0x16c9a118 : "rpc_s_ss_invalid_codeset_tag",
	0x16c9a119 : "rpc_s_mgmt_bad_type",
	0x16c9a11a : "rpc_s_ss_invalid_char_input",
	0x16c9a11b : "rpc_s_ss_short_conv_buffer",
	0x16c9a11c : "rpc_s_ss_iconv_error",
	0x16c9a11d : "rpc_s_ss_no_compat_codeset",
	0x16c9a11e : "rpc_s_ss_no_compat_charsets",
	0x16c9a11f : "dce_cs_c_ok",
	0x16c9a120 : "dce_cs_c_unknown",
	0x16c9a121 : "dce_cs_c_notfound",
	0x16c9a122 : "dce_cs_c_cannot_open_file",
	0x16c9a123 : "dce_cs_c_cannot_read_file",
	0x16c9a124 : "dce_cs_c_cannot_allocate_memory",
	0x16c9a125 : "rpc_s_ss_cleanup_failed",
	0x16c9a126 : "rpc_svc_desc_general",
	0x16c9a127 : "rpc_svc_desc_mutex",
	0x16c9a128 : "rpc_svc_desc_xmit",
	0x16c9a129 : "rpc_svc_desc_recv",
	0x16c9a12a : "rpc_svc_desc_dg_state",
	0x16c9a12b : "rpc_svc_desc_cancel",
	0x16c9a12c : "rpc_svc_desc_orphan",
	0x16c9a12d : "rpc_svc_desc_cn_state",
	0x16c9a12e : "rpc_svc_desc_cn_pkt",
	0x16c9a12f : "rpc_svc_desc_pkt_quotas",
	0x16c9a130 : "rpc_svc_desc_auth",
	0x16c9a131 : "rpc_svc_desc_source",
	0x16c9a132 : "rpc_svc_desc_stats",
	0x16c9a133 : "rpc_svc_desc_mem",
	0x16c9a134 : "rpc_svc_desc_mem_type",
	0x16c9a135 : "rpc_svc_desc_dg_pktlog",
	0x16c9a136 : "rpc_svc_desc_thread_id",
	0x16c9a137 : "rpc_svc_desc_timestamp",
	0x16c9a138 : "rpc_svc_desc_cn_errors",
	0x16c9a139 : "rpc_svc_desc_conv_thread",
	0x16c9a13a : "rpc_svc_desc_pid",
	0x16c9a13b : "rpc_svc_desc_atfork",
	0x16c9a13c : "rpc_svc_desc_cma_thread",
	0x16c9a13d : "rpc_svc_desc_inherit",
	0x16c9a13e : "rpc_svc_desc_dg_sockets",
	0x16c9a13f : "rpc_svc_desc_timer",
	0x16c9a140 : "rpc_svc_desc_threads",
	0x16c9a141 : "rpc_svc_desc_server_call",
	0x16c9a142 : "rpc_svc_desc_nsi",
	0x16c9a143 : "rpc_svc_desc_dg_pkt",
	0x16c9a144 : "rpc_m_cn_ill_state_trans_sa",
	0x16c9a145 : "rpc_m_cn_ill_state_trans_ca",
	0x16c9a146 : "rpc_m_cn_ill_state_trans_sg",
	0x16c9a147 : "rpc_m_cn_ill_state_trans_cg",
	0x16c9a148 : "rpc_m_cn_ill_state_trans_sr",
	0x16c9a149 : "rpc_m_cn_ill_state_trans_cr",
	0x16c9a14a : "rpc_m_bad_pkt_type",
	0x16c9a14b : "rpc_m_prot_mismatch",
	0x16c9a14c : "rpc_m_frag_toobig",
	0x16c9a14d : "rpc_m_unsupp_stub_rtl_if",
	0x16c9a14e : "rpc_m_unhandled_callstate",
	0x16c9a14f : "rpc_m_call_failed",
	0x16c9a150 : "rpc_m_call_failed_no_status",
	0x16c9a151 : "rpc_m_call_failed_errno",
	0x16c9a152 : "rpc_m_call_failed_s",
	0x16c9a153 : "rpc_m_call_failed_c",
	0x16c9a154 : "rpc_m_errmsg_toobig",
	0x16c9a155 : "rpc_m_invalid_srchattr",
	0x16c9a156 : "rpc_m_nts_not_found",
	0x16c9a157 : "rpc_m_invalid_accbytcnt",
	0x16c9a158 : "rpc_m_pre_v2_ifspec",
	0x16c9a159 : "rpc_m_unk_ifspec",
	0x16c9a15a : "rpc_m_recvbuf_toosmall",
	0x16c9a15b : "rpc_m_unalign_authtrl",
	0x16c9a15c : "rpc_m_unexpected_exc",
	0x16c9a15d : "rpc_m_no_stub_data",
	0x16c9a15e : "rpc_m_eventlist_full",
	0x16c9a15f : "rpc_m_unk_sock_type",
	0x16c9a160 : "rpc_m_unimp_call",
	0x16c9a161 : "rpc_m_invalid_seqnum",
	0x16c9a162 : "rpc_m_cant_create_uuid",
	0x16c9a163 : "rpc_m_pre_v2_ss",
	0x16c9a164 : "rpc_m_dgpkt_pool_corrupt",
	0x16c9a165 : "rpc_m_dgpkt_bad_free",
	0x16c9a166 : "rpc_m_lookaside_corrupt",
	0x16c9a167 : "rpc_m_alloc_fail",
	0x16c9a168 : "rpc_m_realloc_fail",
	0x16c9a169 : "rpc_m_cant_open_file",
	0x16c9a16a : "rpc_m_cant_read_addr",
	0x16c9a16b : "rpc_svc_desc_libidl",
	0x16c9a16c : "rpc_m_ctxrundown_nomem",
	0x16c9a16d : "rpc_m_ctxrundown_exc",
	0x16c9a16e : "rpc_s_fault_codeset_conv_error",
	0x16c9a16f : "rpc_s_no_call_active",
	0x16c9a170 : "rpc_s_cannot_support",
	0x16c9a171 : "rpc_s_no_context_available",
}

class DCERPCException(Exception):
	"""
	This is the exception every client should catch regardless of the underlying
	DCERPC Transport used.
	"""
	def __init__(self, error_string=None, error_code=None, packet=None):
		"""
		:param string error_string: A string you want to show explaining the exception. Otherwise the default ones will be used
		:param integer error_code: the error_code if we're using a dictionary with error's descriptions
		:param NDR packet: if successfully decoded, the NDR packet of the response call. This could probably have useful
		information
		"""
		Exception.__init__(self)
		self.packet = packet
		self.error_string = error_string
		if packet is not None:
			try:
				self.error_code = packet['ErrorCode']
			except:
				self.error_code = error_code
		else:
			self.error_code = error_code

	def get_error_code( self ):
		return self.error_code
 
	def get_packet( self ):
		return self.packet

	def __str__( self ):
		key = self.error_code
		if self.error_string is not None:
			return self.error_string
		if key in rpc_status_codes:
			error_msg_short = rpc_status_codes[key]
			return 'DCERPC Runtime Error: code: 0x%x - %s ' % (self.error_code, error_msg_short)
		else:
			return 'DCERPC Runtime Error: unknown error code: 0x%x' % self.error_code

# Context Item
class CtxItem(Structure):
	structure = (
		('ContextID','<H=0'),
		('TransItems','B=0'),
		('Pad','B=0'),
		('AbstractSyntax','20s=""'),
		('TransferSyntax','20s=""'),
	)

class CtxItemResult(Structure):
	structure = (
		('Result','<H=0'),
		('Reason','<H=0'),
		('TransferSyntax','20s=""'),
	)

class SEC_TRAILER(Structure):
	commonHdr = (
		('auth_type', 'B=10'),
		('auth_level','B=0'),
		('auth_pad_len','B=0'),
		('auth_rsvrd','B=0'),
		('auth_ctx_id','<L=747920'),
	)

class MSRPCHeader(Structure):
	_SIZE = 16
	commonHdr = ( 
		('ver_major','B=5'),							  # 0
		('ver_minor','B=0'),							  # 1
		('type','B=0'),								   # 2
		('flags','B=0'),								  # 3
		('representation','<L=0x10'),					 # 4
		('frag_len','<H=self._SIZE+len(auth_data)+(16 if (self["flags"] & 0x80) > 0 else 0)+len(pduData)+len(pad)+len(sec_trailer)'),  # 8
		('auth_len','<H=len(auth_data)'),				 # 10
		('call_id','<L=1'),							   # 12	<-- Common up to here (including this)
	)

	structure = ( 
		('dataLen','_-pduData','self["frag_len"]-self["auth_len"]-self._SIZE-(8 if self["auth_len"] > 0 else 0)'),
		('pduData',':'),								
		('_pad', '_-pad','(4 - ((self._SIZE + (16 if (self["flags"] & 0x80) > 0 else 0) + len(self["pduData"])) & 3) & 3)'),
		('pad', ':'),
		('_sec_trailer', '_-sec_trailer', '8 if self["auth_len"] > 0 else 0'),
		('sec_trailer',':'),
		('auth_dataLen','_-auth_data','self["auth_len"]'),
		('auth_data',':'),
	)

	def __init__(self, data = None, alignment = 0):
		Structure.__init__(self,data, alignment)
		if data is None:
			self['ver_major'] = 5
			self['ver_minor'] = 0
			self['flags'] = PFC_FIRST_FRAG | PFC_LAST_FRAG
			self['type'] = MSRPC_REQUEST
			self.__frag_len_set = 0
			self['auth_len'] = 0
			self['pduData'] = b''
			self['auth_data'] = b''
			self['sec_trailer'] = b''
			self['pad'] = b''

	def get_header_size(self):
		return self._SIZE + (16 if (self["flags"] & PFC_OBJECT_UUID) > 0 else 0)

	def get_packet(self):
		if self['auth_data'] != b'':
			self['auth_len'] = len(self['auth_data'])
		# The sec_trailer structure MUST be 4-byte aligned with respect to 
		# the beginning of the PDU. Padding octets MUST be used to align the 
		# sec_trailer structure if its natural beginning is not already 4-byte aligned
		##self['pad'] = '\xAA' * (4 - ((self._SIZE + len(self['pduData'])) & 3) & 3)

		return self.getData()

class MSRPCRequestHeader(MSRPCHeader):
	_SIZE = 24
	commonHdr = MSRPCHeader.commonHdr + ( 
		('alloc_hint','<L=0'),							# 16
		('ctx_id','<H=0'),								# 20
		('op_num','<H=0'),								# 22
		('_uuid','_-uuid','16 if self["flags"] & 0x80 > 0 else 0' ), # 22
		('uuid',':'),								# 22
	)

	def __init__(self, data = None, alignment = 0):
		MSRPCHeader.__init__(self, data, alignment)
		if data is None:
		   self['type'] = MSRPC_REQUEST
		   self['ctx_id'] = 0
		   self['uuid'] = b''

class MSRPCRespHeader(MSRPCHeader):
	_SIZE = 24
	commonHdr = MSRPCHeader.commonHdr + ( 
		('alloc_hint','<L=0'),						  # 16   
		('ctx_id','<H=0'),							  # 20
		('cancel_count','<B=0'),						# 22
		('padding','<B=0'),							 # 23
	)

	def __init__(self, aBuffer = None, alignment = 0):
		MSRPCHeader.__init__(self, aBuffer, alignment)
		if aBuffer is None:
			self['type'] = MSRPC_RESPONSE
			self['ctx_id'] = 0

class MSRPCBind(Structure):
	_CTX_ITEM_LEN = len(CtxItem())
	structure = ( 
		('max_tfrag','<H=4280'),
		('max_rfrag','<H=4280'),
		('assoc_group','<L=0'),
		('ctx_num','B=0'),
		('Reserved','B=0'),
		('Reserved2','<H=0'),
		('_ctx_items', '_-ctx_items', 'self["ctx_num"]*self._CTX_ITEM_LEN'),
		('ctx_items',':'),
	)
 
	def __init__(self, data = None, alignment = 0):
		Structure.__init__(self, data, alignment)
		if data is None:
			self['max_tfrag'] = 4280
			self['max_rfrag'] = 4280
			self['assoc_group'] = 0
			self['ctx_num'] = 1
			self['ctx_items'] = b''
		self.__ctx_items = []

	def addCtxItem(self, item):
		self.__ctx_items.append(item)
	
	def getData(self):
		self['ctx_num'] = len(self.__ctx_items)
		for i in self.__ctx_items:
			self['ctx_items'] += i.getData()
		return Structure.getData(self)

class MSRPCBindAck(MSRPCHeader):
	_SIZE = 26 # Up to SecondaryAddr
	_CTX_ITEM_LEN = len(CtxItemResult())
	structure = ( 
		('max_tfrag','<H=0'),
		('max_rfrag','<H=0'),
		('assoc_group','<L=0'),
		('SecondaryAddrLen','<H&SecondaryAddr'), 
		('SecondaryAddr','z'),						  # Optional if SecondaryAddrLen == 0
		('PadLen','_-Pad','(4-((self["SecondaryAddrLen"]+self._SIZE) % 4))%4'),
		('Pad',':'),
		('ctx_num','B=0'),
		('Reserved','B=0'),
		('Reserved2','<H=0'),
		('_ctx_items','_-ctx_items','self["ctx_num"]*self._CTX_ITEM_LEN'),
		('ctx_items',':'),
		('_sec_trailer', '_-sec_trailer', '8 if self["auth_len"] > 0 else 0'),
		('sec_trailer',':'),
		('auth_dataLen','_-auth_data','self["auth_len"]'),
		('auth_data',':'),
	)
	def __init__(self, data = None, alignment = 0):
		self.__ctx_items = []
		MSRPCHeader.__init__(self,data,alignment)
		if data is None:
			self['Pad'] = b''
			self['ctx_items'] = b''
			self['sec_trailer'] = b''
			self['auth_data'] = b''

	def getCtxItems(self):
		return self.__ctx_items

	def getCtxItem(self,index):
		return self.__ctx_items[index-1]

	def fromString(self, data):
		Structure.fromString(self,data)
		# Parse the ctx_items
		data = self['ctx_items']
		for i in range(self['ctx_num']):
			item = CtxItemResult(data)
			self.__ctx_items.append(item)
			data = data[len(item):]
			
class MSRPCBindNak(Structure):
	structure = ( 
		('RejectedReason','<H=0'),
		('SupportedVersions',':'),
	)
	def __init__(self, data = None, alignment = 0):
		Structure.__init__(self,data,alignment)
		if data is None:
			self['SupportedVersions'] = b''

class DCERPC:
	# Standard NDR Representation
	NDRSyntax   = uuidtup_to_bin(('8a885d04-1ceb-11c9-9fe8-08002b104860', '2.0'))
	# NDR 64
	NDR64Syntax = uuidtup_to_bin(('71710533-BEBA-4937-8319-B5DBEF9CCC36', '1.0'))
	transfer_syntax =  NDRSyntax


class DCERPC_RawCall(MSRPCRequestHeader):
	def __init__(self, op_num, data = b'', uuid=None):
		MSRPCRequestHeader.__init__(self)
		self['op_num'] = op_num
		self['pduData'] = data
		if uuid is not None:
			self['flags'] |= PFC_OBJECT_UUID
			self['uuid'] = uuid

	def setData(self, data):
		self['pduData'] = data

# 2.2.6 Type Serialization Version 1
class CommonHeader(NDRSTRUCT):
	structure = (
		('Version', UCHAR),
		('Endianness', UCHAR),
		('CommonHeaderLength', USHORT),
		('Filler', ULONG),
	)
	def __init__(self, data = None,isNDR64 = False):
		NDRSTRUCT.__init__(self, data, isNDR64)
		if data is None:
			self['Version'] = 1
			self['Endianness'] = 0x10
			self['CommonHeaderLength'] = 8
			self['Filler'] = 0xcccccccc

class PrivateHeader(NDRSTRUCT):
	structure = (
		('ObjectBufferLength', ULONG),
		('Filler', ULONG),
	)
	def __init__(self, data = None,isNDR64 = False):
		NDRSTRUCT.__init__(self, data, isNDR64)
		if data is None:
			self['Filler'] = 0xcccccccc

class TypeSerialization1(NDRSTRUCT):
	commonHdr = (
		('CommonHeader', CommonHeader),
		('PrivateHeader', PrivateHeader),
	)
	def getData(self, soFar = 0):
		self['PrivateHeader']['ObjectBufferLength'] = len(NDRSTRUCT.getData(self, soFar)) + len(
			NDRSTRUCT.getDataReferents(self, soFar)) - len(self['CommonHeader']) - len(self['PrivateHeader'])
		return NDRSTRUCT.getData(self, soFar)



from aiosmb.commons.connection.proxy import SMBProxyType
from aiosmb.network.tcp import TCPSocket
from aiosmb.network.socks import SocksProxyConnection
from aiosmb.network.multiplexornetwork import MultiplexorProxyConnection



class NetworkSelector:
    def __init__(self):
        pass

    @staticmethod
    async def select(target):
        if target.proxy is None:
            return TCPSocket(target = target)
        elif target.proxy.type in [SMBProxyType.SOCKS5, SMBProxyType.SOCKS5_SSL, SMBProxyType.SOCKS4, SMBProxyType.SOCKS4_SSL]:
            return SocksProxyConnection(target = target)

        elif target.proxy.type in [SMBProxyType.MULTIPLEXOR, SMBProxyType.MULTIPLEXOR_SSL]:
            mpc = MultiplexorProxyConnection(target)
            socks_proxy = await mpc.connect()
            return socks_proxy

        return None


import ipaddress
import enum
from urllib.parse import urlparse, parse_qs

from asysocks.common.clienturl import SocksClientURL 


def stru(x):
	return str(x).upper()

class SMBProxySecretType(enum.Enum):
	NONE = 'NONE'
	PLAIN = 'PLAIN'

class SMBProxyType(enum.Enum):
	SOCKS4 = 'SOCKS4'
	SOCKS4_SSL = 'SOCKS4_SSL'
	SOCKS5 = 'SOCKS5'
	SOCKS5_SSL = 'SOCKS5_SSL'
	MULTIPLEXOR = 'MULTIPLEXOR'
	MULTIPLEXOR_SSL = 'MULTIPLEXOR_SSL'

multiplexorproxyurl_param2var = {
	'type' : ('version', [stru, SMBProxyType]),
	'host' : ('server_ip', [str]),
	'port' : ('server_port', [int]),
	'timeout': ('timeout', [int]),
	'user' : ('username', [str]),
	'pass' : ('password', [str]),
	#'authtype' : ('authtype', [SOCKS5Method]),
	'agentid' : ('agentid', [str]),
	'domain' : ('domain', [str])

}


class SMBProxy:
	def __init__(self):
		self.type = None
		self.target = None
		self.auth   = None

	@staticmethod
	def from_params(url_str):
		proxy = SMBProxy()
		url = urlparse(url_str)
		if url.query is None:
			return None

		query = parse_qs(url.query)
		if 'proxytype' not in query and 'sametype' not in query:
			return None
		
		proxy.type = SMBProxyType(query['proxytype'][0].upper())
		if proxy.type in [SMBProxyType.SOCKS4, SMBProxyType.SOCKS4_SSL, SMBProxyType.SOCKS5, SMBProxyType.SOCKS5_SSL]:
			cu = SocksClientURL.from_params(url_str)
			cu.endpoint_port = 445
		else:
			cu = SocksClientURL.from_params(url_str)
		
		proxy.target = cu.get_target()
		return proxy

	def __str__(self):
		t = '==== SMBProxy ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t

class SMBMultiplexorProxy:
	def __init__(self):
		self.ip = None
		self.port = None
		self.timeout = 10
		self.type = SMBProxyType.MULTIPLEXOR
		self.username = None
		self.password = None
		self.domain = None
		self.agent_id = None
	
	def sanity_check(self):
		if self.ip is None:
			raise Exception('MULTIPLEXOR server IP is missing!')
		if self.port is None:
			raise Exception('MULTIPLEXOR server port is missing!')
		if self.agent_id is None:
				raise Exception('MULTIPLEXOR proxy requires agentid to be set!')

	@staticmethod
	def from_params(url_str):
		res = SMBMultiplexorProxy()
		url = urlparse(url_str)
		res.endpoint_ip = url.hostname
		if url.port:
			res.endpoint_port = int(url.port)
		if url.query is not None:
			query = parse_qs(url.query)

			for k in query:
				if k.startswith('proxy'):
					if k[5:] in multiplexorproxyurl_param2var:

						data = query[k][0]
						for c in multiplexorproxyurl_param2var[k[5:]][1]:
							#print(c)
							data = c(data)

						setattr(
							res, 
							multiplexorproxyurl_param2var[k[5:]][0], 
							data
						)
		res.sanity_check()


#class SMBProxy:
#	def __init__(self, 
#				ip = None,
#				port = None,
#				timeout = 5,
#				proxy_type = None,
#				username = None,
#				domain = None,
#				secret = None,
#				secret_type = SMBProxySecretType.NONE,
#				agent_id = None):
#		self.ip = ip
#		self.port = port
#		self.timeout = timeout
#		self.type = proxy_type
#		self.username = username
#		self.domain = domain
#		self.secret = secret
#		self.secret_type = secret_type
#		self.agent_id = agent_id #used by multiplexor only
#		self.settings = {}
#	
#	@staticmethod
#	def from_url(url_str):
#		proxy = SMBProxy()
#		url_e = urlparse(url_str)
#		if url_e.query is None:
#			return None
#		
#		query = parse_qs(url_e.query)
#		if 'proxytype' not in query and 'sametype' not in query:
#			return None
#
#		for k in query:
#			if k.startswith('proxy') or k.startswith('same'):
#				key = 'same'
#				if k.startswith('proxy'):
#					key = 'proxy'
#				
#				if k == key+'type':
#					proxy.type = SMBProxyType(query[k][0].replace('-','_').upper())
#				elif k == key+'user':
#					proxy.username = query[k][0]
#				elif k == key+'domain':
#					proxy.domain = query[k][0]
#				elif k == key+'pass':
#					proxy.secret = query[k][0]
#				elif k == key+'host':
#					proxy.ip = query[k][0]
#				elif k == key+'port':
#					proxy.port = int(query[k][0])
#				elif k == key+'timeout':
#					proxy.timeout = int(query[k][0])
#				elif k == key+'agentid':
#					proxy.agent_id = query[k][0]
#				else:
#					proxy.settings[k[len(key):]] = query[k] #the result is a list for each entry because this preprocessor is not aware which elements should be lists!
#		
#		if proxy.secret is not None:
#			proxy.secret_type = SMBProxySecretType.PLAIN
#		else:
#			proxy.secret_type = SMBProxySecretType.NONE
#
#		if proxy.ip is None:
#			raise Exception('Proxy server specified with missing servber address! (add proxyhost to params)')
#		
#		if proxy.type in [SMBProxyType.MULTIPLEXOR_SSL, SMBProxyType.MULTIPLEXOR]:
#			if proxy.agent_id is None:
#				raise Exception('MULTIPLEXOR proxy requires agentid to be set!')
#
#		return proxy
#
#	
#	@staticmethod
#	def from_connection_string(s):
#		"""
#		URL format required
#		socks5://user:password@ipaddress:port
#		socks5+ssl://user:password@ipaddress:port
#		multiplexor://user:password@ipaddress:port/agentid
#		"""
#		st = SMBProxy()
#		o = urlparse(s)
#
#		st.type = SMBProxyType(o.scheme.upper().replace('+','_'))
#		st.ip = o.hostname
#		st.username = o.username
#		st.secret = o.password
#		st.port = int(o.port)
#		if st.secret is not None:
#			st.secret_type = SMBProxySecretType.PLAIN
#		
#		if st.type in [SMBProxyType.MULTIPLEXOR, SMBProxyType.MULTIPLEXOR_SSL]:
#			st.agent_id = o.path.replace('/','')
#		
#		return st
#		
#	def __str__(self):
#		t = '==== SMBProxy ====\r\n'
#		for k in self.__dict__:
#			t += '%s: %s\r\n' % (k, self.__dict__[k])
#			
#		return t
#		
		
def test():
	t = ['socks5://10.10.10.1',
			'socks5+ssl://10.10.10.1',
			'socks5+ssl://admin:password@10.10.10.1',
			'multiplexor+ssl://admin:password@10.10.10.1/alma',
	]
	for x in t:
		s = SMBProxy.from_connection_string(x)
		print(str(s))

	
if __name__ == '__main__':
	test()


import io
import enum

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/8a622ecb-ffee-41b9-b4c4-83ff2d3aba1b
class TREE_DISCONNECT_REQ:
	def __init__(self):
		self.StructureSize = 4
		self.Reserved = 0
		
	def to_bytes(self):
		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff):
		return TREE_DISCONNECT_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = TREE_DISCONNECT_REQ()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 4
		msg.Reserved  = int.from_bytes(buff.read(2), byteorder='little')
		return msg

	def __repr__(self):
		t = '==== SMB2 TREE DISCONNECT REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		return t
		
		
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/aeac92de-8db3-48f8-a8b7-bfee28b9fd9e
class TREE_DISCONNECT_REPLY:
	def __init__(self):
		self.StructureSize = 4
		self.Reserved = 0
		
	def to_bytes(self):
		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(2, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff):
		return TREE_DISCONNECT_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = TREE_DISCONNECT_REPLY()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 4
		msg.Reserved  = int.from_bytes(buff.read(2), byteorder='little')
		return msg

	def __repr__(self):
		t = '==== SMB2 TREE DISCONNECT REPLY ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		return t


import io
import enum

# TODO: additional parsing for the error context!

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/d4da8b67-c180-47e3-ba7a-d24214ac4aaa
class ERROR_REPLY:
	def __init__(self):
		self.StructureSize = 9
		self.ErrorContextCount = 0
		self.Reserved = 0
		self.ByteCount = 0
		
		self.ErrorData = None
		
	def to_bytes(self):
		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.ErrorContextCount.to_bytes(1, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(1, byteorder='little', signed = False)
		t += self.ByteCount.to_bytes(4, byteorder='little', signed = False)
		if self.ByteCount > 0:
			t += self.ErrorData.to_bytes()
		return t

	@staticmethod
	def from_bytes(bbuff):
		return ERROR_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = ERROR_REPLY()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 9
		msg.ErrorContextCount  = int.from_bytes(buff.read(1), byteorder='little')
		msg.Reserved  = int.from_bytes(buff.read(1), byteorder='little')
		msg.ByteCount  = int.from_bytes(buff.read(4), byteorder='little')
		if msg.ByteCount > 0:
			msg.ErrorData = buff.read(msg.ByteCount)
		return msg

	def __repr__(self):
		t = '==== SMB2 ERROR ====\r\n'
		return t


# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Author: Alberto Solino (beto@coresecurity.com)
#
# Description:
#   RFC 4493 implementation (https://www.ietf.org/rfc/rfc4493.txt)
#   RFC 4615 implementation (https://www.ietf.org/rfc/rfc4615.txt)
#
#   NIST SP 800-108 Section 5.1, with PRF HMAC-SHA256 implementation
#   (https://tools.ietf.org/html/draft-irtf-cfrg-kdf-uses-00#ref-SP800-108)
#
#   [MS-LSAD] Section 5.1.2
#   [MS-SAMR] Section 2.2.11.1.1

from struct import pack, unpack

import hashlib
import hmac

from aiosmb.crypto.symmetric import AES
from aiosmb.crypto.BASE import cipherMODE

def KDF_CounterMode(KI, Label, Context, L):
# Implements NIST SP 800-108 Section 5.1, with PRF HMAC-SHA256
# https://tools.ietf.org/html/draft-irtf-cfrg-kdf-uses-00#ref-SP800-108
# Fixed values:
#  1. h - The length of the output of the PRF in bits, and
#  2. r - The length of the binary representation of the counter i.
# Input: KI, Label, Context, and L.
# Process:
#  1. n := [L/h]
#  2. If n > 2r-1, then indicate an error and stop.
#  3. result(0):= empty .
#  4. For i = 1 to n, do
#    a. K(i) := PRF (KI, [i]2 || Label || 0x00 || Context || [L]2)
#    b. result(i) := result(i-1) || K(i).
#  5. Return: KO := the leftmost L bits of result(n).
	h = 256
	r = 32

	n = L // h

	if n == 0:
		n = 1

	if n > (pow(2,r)-1):
		raise Exception("Error computing KDF_CounterMode")

	result = b''
	K      = b''

	for i in range(1,n+1):
	   input = pack('>L', i) + Label + b'\x00' + Context + pack('>L',L)
	   K = hmac.new(KI, input, hashlib.sha256).digest()
	   result = result + K

	return result[:(L//8)]

def AES_CMAC(K, M, length):

#   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#   +                   Algorithm AES-CMAC                              +
#   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#   +                                                                   +
#   +   Input    : K    ( 128-bit key )                                 +
#   +            : M    ( message to be authenticated )                 +
#   +            : len  ( length of the message in octets )             +
#   +   Output   : T    ( message authentication code )                 +
#   +                                                                   +
#   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#   +   Constants: const_Zero is 0x00000000000000000000000000000000     +
#   +              const_Bsize is 16                                    +
#   +                                                                   +
#   +   Variables: K1, K2 for 128-bit subkeys                           +
#   +              M_i is the i-th block (i=1..ceil(len/const_Bsize))   +
#   +              M_last is the last block xor-ed with K1 or K2        +
#   +              n      for number of blocks to be processed          +
#   +              r      for number of octets of last block            +
#   +              flag   for denoting if last block is complete or not +
#   +                                                                   +
#   +   Step 1.  (K1,K2) := Generate_Subkey(K);                         +
#   +   Step 2.  n := ceil(len/const_Bsize);                            +
#   +   Step 3.  if n = 0                                               +
#   +            then                                                   +
#   +                 n := 1;                                           +
#   +                 flag := false;                                    +
#   +            else                                                   +
#   +                 if len mod const_Bsize is 0                       +
#   +                 then flag := true;                                +
#   +                 else flag := false;                               +
#   +                                                                   +
#   +   Step 4.  if flag is true                                        +
#   +            then M_last := M_n XOR K1;                             +
#   +            else M_last := padding(M_n) XOR K2;                    +
#   +   Step 5.  X := const_Zero;                                       +
#   +   Step 6.  for i := 1 to n-1 do                                   +
#   +                begin                                              +
#   +                  Y := X XOR M_i;                                  +
#   +                  X := AES-128(K,Y);                               +
#   +                end                                                +
#   +            Y := M_last XOR X;                                     +
#   +            T := AES-128(K,Y);                                     +
#   +   Step 7.  return T;                                              +
#   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	const_Bsize = 16
	const_Zero  = bytearray(16)

	AES_128= AES(K, cipherMODE.ECB)
	M      = bytearray(M[:length])
	K1, K2 = Generate_Subkey(K)
	n      = len(M)//const_Bsize

	if n == 0:
		n = 1
		flag = False
	else:
		if (length % const_Bsize) == 0:
			flag = True
		else:
			n += 1
			flag = False

	M_n = M[(n-1)*const_Bsize:]
	if flag is True:
		M_last = XOR_128(M_n,K1)
	else:
		M_last = XOR_128(PAD(M_n),K2)

	X = const_Zero
	for i in range(n-1):
		M_i = M[(i)*const_Bsize:][:16]
		Y   = XOR_128(X, M_i)
		X   = bytearray(AES_128.encrypt(bytes(Y)))
	Y = XOR_128(M_last, X)
	T = AES_128.encrypt(bytes(Y))

	return T


def Generate_Subkey(K):

#   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#   +                    Algorithm Generate_Subkey                      +
#   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#   +                                                                   +
#   +   Input    : K (128-bit key)                                      +
#   +   Output   : K1 (128-bit first subkey)                            +
#   +              K2 (128-bit second subkey)                           +
#   +-------------------------------------------------------------------+
#   +                                                                   +
#   +   Constants: const_Zero is 0x00000000000000000000000000000000     +
#   +              const_Rb   is 0x00000000000000000000000000000087     +
#   +   Variables: L          for output of AES-128 applied to 0^128    +
#   +                                                                   +
#   +   Step 1.  L := AES-128(K, const_Zero);                           +
#   +   Step 2.  if MSB(L) is equal to 0                                +
#   +            then    K1 := L << 1;                                  +
#   +            else    K1 := (L << 1) XOR const_Rb;                   +
#   +   Step 3.  if MSB(K1) is equal to 0                               +
#   +            then    K2 := K1 << 1;                                 +
#   +            else    K2 := (K1 << 1) XOR const_Rb;                  +
#   +   Step 4.  return K1, K2;                                         +
#   +                                                                   +
#   +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	AES_128 = AES(K, cipherMODE.ECB)

	L = AES_128.encrypt(bytes(bytearray(16)))

	LHigh = unpack('>Q',L[:8])[0]
	LLow  = unpack('>Q',L[8:])[0]

	K1High = ((LHigh << 1) | ( LLow >> 63 )) & 0xFFFFFFFFFFFFFFFF
	K1Low  = (LLow << 1) & 0xFFFFFFFFFFFFFFFF

	if (LHigh >> 63):
		K1Low ^= 0x87

	K2High = ((K1High << 1) | (K1Low >> 63)) & 0xFFFFFFFFFFFFFFFF
	K2Low  = ((K1Low << 1)) & 0xFFFFFFFFFFFFFFFF

	if (K1High >> 63):
		K2Low ^= 0x87

	K1 = bytearray(pack('>QQ', K1High, K1Low))
	K2 = bytearray(pack('>QQ', K2High, K2Low))

	return K1, K2

def XOR_128(N1,N2):

	J = bytearray()
	for i in range(len(N1)):
		#J.append(indexbytes(N1,i) ^ indexbytes(N2,i))
		J.append(N1[i] ^ N2[i])
	return J

def PAD(N):
	padLen = 16-len(N)
	return  N + b'\x80' + b'\x00'*(padLen-1)


import enum


class SMB2HeaderFlag(enum.IntFlag):
	SMB2_FLAGS_SERVER_TO_REDIR    = 0x00000001  # When set, indicates the message is a response rather than a request. This MUST be set on responses sent from the server to the client, and MUST NOT be set on requests sent from the client to the server.
	SMB2_FLAGS_ASYNC_COMMAND      = 0x00000002  # When set, indicates that this is an ASYNC SMB2 header. Always set for headers of the form described in this section.
	SMB2_FLAGS_RELATED_OPERATIONS = 0x00000004
	SMB2_FLAGS_SIGNED             = 0x00000008  # When set, indicates that this packet has been signed. The use of this flag is as specified in section 3.1.5.1.
	SMB2_FLAGS_PRIORITY_MASK      = 0x00000070  # This flag is only valid for the SMB 3.1.1 dialect. It is a mask for the requested I/O priority of the request, and it MUST be a value in the range 0 to 7.
	SMB2_FLAGS_DFS_OPERATIONS     = 0x10000000  # When set, indicates that this command is a Distributed File System (DFS) operation. The use of this flag is as specified in section 3.3.5.9.
	SMB2_FLAGS_REPLAY_OPERATION   = 0x20000000






import os
import struct
import hmac
import copy

from aiosmb.commons.connection.credential import SMBNTLMCredential
from aiosmb.commons.serverinfo import NTLMServerInfo
from aiosmb.authentication.ntlm.templates.server import NTLMServerTemplates
from aiosmb.authentication.ntlm.templates.client import NTLMClientTemplates
from aiosmb.authentication.ntlm.structures.negotiate_flags import NegotiateFlags
from aiosmb.authentication.ntlm.structures.version import Version
from aiosmb.authentication.ntlm.structures.ntlmssp_message_signature import NTLMSSP_MESSAGE_SIGNATURE
from aiosmb.authentication.ntlm.structures.ntlmssp_message_signature_noext import NTLMSSP_MESSAGE_SIGNATURE_NOEXT
from aiosmb.authentication.ntlm.messages.negotiate import NTLMNegotiate
from aiosmb.authentication.ntlm.messages.challenge import NTLMChallenge
from aiosmb.authentication.ntlm.messages.authenticate import NTLMAuthenticate
from aiosmb.authentication.ntlm.creds_calc import *
from aiosmb.crypto.symmetric import RC4
		

class NTLMHandlerSettings:
	def __init__(self, credential, mode = 'CLIENT', template_name = 'Windows10_15063', custom_template = None):
		self.credential = credential
		self.mode = mode
		self.template_name = template_name
		self.custom_template = custom_template #for custom templates, must be dict
		
		self.template = None
		self.ntlm_downgrade = False
		
		self.construct_message_template()
		
	def construct_message_template(self):
		if self.mode.upper() == 'MANUAL':
			return
			
		if not self.template_name:
			if not self.custom_template:
				raise Exception('No NTLM tamplate specified!')
			
			self.template = self.custom_template
		
		if self.mode.upper() == 'SERVER':
			if self.template_name in NTLMServerTemplates:
				self.template = NTLMServerTemplates[self.template_name]
			else:
				raise Exception('No NTLM server template found with name %s' % self.template_name)
	
		else:
			if self.template_name in NTLMClientTemplates:
				self.template = NTLMClientTemplates[self.template_name]
				if 'ntlm_downgrade' in self.template:
					self.ntlm_downgrade = self.template['ntlm_downgrade']
			else:
					raise Exception('No NTLM server template found with name %s' % self.template_name)
		

class NTLMAUTHHandler:
	def __init__(self, settings):
		self.settings = settings #NTLMHandlerSettings		
		
		self.mode = None
		self.flags = None
		self.challenge = None
		
		self.ntlmNegotiate     = None #ntlm Negotiate message from client
		self.ntlmChallenge     = None #ntlm Challenge message to client
		self.ntlmAuthenticate  = None #ntlm Authenticate message from client
		
		self.ntlmNegotiate_raw     = None #message as bytes, as it's recieved/sent
		self.ntlmChallenge_raw     = None #message as bytes, as it's recieved/sent
		self.ntlmAuthenticate_raw  = None #message as bytes, as it's recieved/sent

		
		self.EncryptedRandomSessionKey = None
		self.RandomSessionKey = None
		self.SessionBaseKey = None
		self.KeyExchangeKey = None
		
		self.SignKey_client = None
		self.SealKey_client = None
		self.SignKey_server = None
		self.SealKey_server = None
		
		self.iteration_cnt = 0
		self.ntlm_credentials = None
		self.timestamp = None #used in unittest only!
		self.extra_info = None
		self.setup()

	def setup(self):
		self.mode = self.settings.mode
		if self.mode.upper() == 'MANUAL':
			#for passign the messages automatically with the sessionbasekey, the using this class for sign and seal
			return
		
		if 'challenge' not in self.settings.template:
			self.challenge = os.urandom(8)
		else:
			self.challenge = self.settings.template['challenge']
		self.flags = self.settings.template['flags']
		if 'session_key' in self.settings.template:
			self.RandomSessionKey = self.settings.template['session_key']
		
		self.timestamp = self.settings.template.get('timestamp') #used in unittest only!
		
					
		if self.mode.upper() == 'SERVER':
			version    = self.settings.template['version']
			targetName = self.settings.template['targetname']
			targetInfo = self.settings.template['targetinfo']

			self.ntlmChallenge = NTLMChallenge.construct(challenge = self.challenge, targetName = targetName, targetInfo = targetInfo, version = version, flags = self.flags)
		
		#else:
		#	domainname = self.settings.template['domain_name']
		#	workstationname = self.settings.template['workstation_name']
		#	version = self.settings.template.get('version')
			
	def load_negotiate(self, data):
		self.ntlmNegotiate = NTLMNegotiate.from_bytes(data)
	
	def load_challenge(self, data):
		self.ntlmChallenge = NTLMChallenge.from_bytes(data)
		
	def load_authenticate(self, data):
		self.ntlmAuthenticate = NTLMAuthenticate.from_bytes(data)
		
	def load_sessionkey(self, data):
		self.RandomSessionKey = data
		self.setup_crypto()
	
	def set_sign(self, tf = True):
		if tf == True:
			self.flags |= NegotiateFlags.NEGOTIATE_SIGN
		else:
			self.flags &= ~NegotiateFlags.NEGOTIATE_SIGN
			
	def set_seal(self, tf = True):
		if tf == True:
			self.flags |= NegotiateFlags.NEGOTIATE_SEAL
		else:
			self.flags &= ~NegotiateFlags.NEGOTIATE_SEAL
			
	def set_version(self, tf = True):
		if tf == True:
			self.flags |= NegotiateFlags.NEGOTIATE_VERSION
		else:
			self.flags &= ~NegotiateFlags.NEGOTIATE_VERSION
			
	def is_extended_security(self):
		return NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY in self.ntlmChallenge.NegotiateFlags
	
	def get_extra_info(self):
		self.extra_info = NTLMServerInfo.from_challenge(self.ntlmChallenge)
		return self.extra_info
		
	def MAC(self, handle, signingKey, seqNum, message):
		if self.is_extended_security() == True:
			msg = NTLMSSP_MESSAGE_SIGNATURE()
			if NegotiateFlags.NEGOTIATE_KEY_EXCH in self.ntlmChallenge.NegotiateFlags:
				tt = struct.pack('<i', seqNum) + message
				t = hmac_md5(signingKey)
				t.update(tt)
				
				msg.Checksum = handle(t.digest()[:8])
				msg.SeqNum = seqNum
				seqNum += 1
			else:
				t = hmac_md5(signingKey)
				t.update(struct.pack('<i',seqNum)+message)
				msg.Checksum = t.digest()[:8]
				msg.SeqNum = seqNum
				seqNum += 1
				
		else:
			raise Exception('Not implemented!')
			#t = struct.pack('<I',binascii.crc32(message)& 0xFFFFFFFF)
			#randompad = 0
			#msg = NTLMSSP_MESSAGE_SIGNATURE_NOEXT()
			#msg.RandomPad = handle(struct.pack('<I',randompad))
			#msg.Checksum = struct.unpack('<I',handle(messageSignature['Checksum']))[0]
			
		return msg.to_bytes()
				
		
	def SEAL(self, signingKey, sealingKey, messageToSign, messageToEncrypt, seqNum, cipher_encrypt):
		sealedMessage = cipher_encrypt(messageToEncrypt)
		signature = self.MAC(cipher_encrypt, signingKey, seqNum, messageToSign)
		return sealedMessage, signature
		
	def SIGN(self, signingKey, message, seqNum, cipher_encrypt):
		return self.MAC(cipher_encrypt, signingKey, seqNum, message)
		
	def calc_sealkey(self, mode = 'Client'):
		if NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY in self.ntlmChallenge.NegotiateFlags:
			if NegotiateFlags.NEGOTIATE_128 in self.ntlmChallenge.NegotiateFlags:
				sealkey = self.RandomSessionKey
			elif NegotiateFlags.NEGOTIATE_56 in self.ntlmChallenge.NegotiateFlags:
				sealkey = self.RandomSessionKey[:7]
			else:
				sealkey = self.RandomSessionKey[:5]
				
			if mode == 'Client':
				md5 = hashlib.new('md5')
				md5.update(sealkey + b'session key to client-to-server sealing key magic constant\x00')
				sealkey = md5.digest()
			else:
				md5 = hashlib.new('md5')
				md5.update(sealkey + b'session key to server-to-client sealing key magic constant\x00')
				sealkey = md5.digest()
				
		elif NegotiateFlags.NEGOTIATE_56 in self.ntlmChallenge.NegotiateFlags:
			sealkey = self.RandomSessionKey[:7] + b'\xa0'
		else:
			sealkey = self.RandomSessionKey[:5] + b'\xe5\x38\xb0'
			
		if mode == 'Client':
			self.SealKey_client = sealkey
		else:
			self.SealKey_server = sealkey
			
		return sealkey
		
	def calc_signkey(self, mode = 'Client'):
		if NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY in self.ntlmChallenge.NegotiateFlags:
			if mode == 'Client':
				md5 = hashlib.new('md5')
				md5.update(self.RandomSessionKey + b"session key to client-to-server signing key magic constant\x00")
				signkey = md5.digest()
			else:
				md5 = hashlib.new('md5')
				md5.update(self.RandomSessionKey + b"session key to server-to-client signing key magic constant\x00")
				signkey = md5.digest()
		else:
			signkey = None
			
		if mode == 'Client':
			self.SignKey_client = signkey
		else:
			self.SignKey_server = signkey
		
		return signkey
		
	def get_session_key(self):
		return self.RandomSessionKey
		
	def get_sealkey(self, mode = 'Client'):
		if mode == 'Client':
			return self.SealKey_client
		else:
			return self.SealKey_server
			
	def get_signkey(self, mode = 'Client'):
		if mode == 'Client':
			return self.SignKey_client
		else:
			return self.SignKey_server
		
	def setup_crypto(self):
		if not self.RandomSessionKey:
			self.RandomSessionKey = os.urandom(16)
		
		if self.mode.upper() != 'MANUAL':
			#this check is here to provide the option to load the messages + the sessionbasekey manually
			#then you will be able to use the sign and seal functions provided by this class
			self.SessionBaseKey = self.ntlm_credentials.SessionBaseKey
		
			rc4 = RC4(self.KeyExchangeKey)
			self.EncryptedRandomSessionKey = rc4.encrypt(self.RandomSessionKey)
		
		self.calc_sealkey('Client')
		self.calc_sealkey('Server')
		self.calc_signkey('Client')
		self.calc_signkey('Server')

	async def authenticate(self, authData, flags = None, seq_number = 0, is_rpc = False):
		if self.mode.upper() == 'SERVER':
			if self.ntlmNegotiate is None:
				###parse client NTLMNegotiate message
				self.ntlmNegotiate = NTLMNegotiate.from_bytes(authData)
				return self.ntlmChallenge.to_bytes(), True 

			elif self.ntlmAuthenticate is None:
				self.ntlmAuthenticate = NTLMAuthenticate.from_bytes(authData, self.use_NTLMv2)
				creds = NTLMcredential.construct(self.ntlmNegotiate, self.ntlmChallenge, self.ntlmAuthenticate)
				print(creds)

				# TODO: check when is sessionkey needed and check when is singing needed, and calculate the keys!
				# self.calc_SessionBaseKey()
				# self.calc_KeyExchangeKey()
				auth_credential = creds[0]
				#self.SessionBaseKey = auth_credential.calc_session_base_key()
				#self.calc_key_exchange_key()

				if auth_credential.verify(self.credential):
					return AuthResult.FAIL, auth_credential
				else:
					return AuthResult.FAIL, auth_credential

			else:
				raise Exception('Too many calls to do_AUTH function!')
				
		elif self.mode.upper() == 'CLIENT':
			if self.iteration_cnt == 0:
				if authData is not None:
					raise Exception('First call as client MUST be with empty data!')
					
				if is_rpc == True:
					#rpc (unknow reason) reqauires seal to be set, otherwise it will fail to authenticate
					self.set_seal()
				
				self.iteration_cnt += 1
				#negotiate message was already calulcated in setup
				self.ntlmNegotiate = NTLMNegotiate.construct(self.flags, domainname = self.settings.template['domain_name'], workstationname = self.settings.template['workstation_name'], version = self.settings.template.get('version'))			
				self.ntlmNegotiate_raw = self.ntlmNegotiate.to_bytes()
				return self.ntlmNegotiate_raw, True
				
			else:
				#server challenge incoming
				self.ntlmChallenge_raw = authData
				self.ntlmChallenge = NTLMChallenge.from_bytes(authData)
				
				
				if is_rpc == True:
					#rpc (unknow reason) reqauires seal to be set, otherwise it will fail to authenticate
					self.set_seal()
				
				##################self.flags = self.ntlmChallenge.NegotiateFlags
				
				#we need to calculate the response based on the credential and the settings flags
				if self.settings.ntlm_downgrade == True:
					#NTLMv1 authentication
					# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/464551a8-9fc4-428e-b3d3-bc5bfb2e73a5
					
					#check if we authenticate as guest
					if self.settings.credential.is_guest == True:
						lmresp = LMResponse()
						lmresp.Response = b'\x00'
						self.ntlmAuthenticate = NTLMAuthenticate.construct(self.flags, lm_response= lmresp)
						return self.ntlmAuthenticate.to_bytes(), False
						
					if self.flags & NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY:
						#Extended auth!
						self.ntlm_credentials = netntlm_ess.construct(self.ntlmChallenge.ServerChallenge, self.challenge, self.settings.credential)
						
						self.KeyExchangeKey = self.ntlm_credentials.calc_key_exchange_key()						
						self.setup_crypto()
						
						self.ntlmAuthenticate = NTLMAuthenticate.construct(self.flags, lm_response= self.ntlm_credentials.LMResponse, nt_response = self.ntlm_credentials.NTResponse, version = self.ntlmNegotiate.Version, encrypted_session = self.EncryptedRandomSessionKey)
					else:
						self.ntlm_credentials = netntlm.construct(self.ntlmChallenge.ServerChallenge, self.settings.credential)
						
						self.KeyExchangeKey = self.ntlm_credentials.calc_key_exchange_key(with_lm = self.flags & NegotiateFlags.NEGOTIATE_LM_KEY, non_nt_session_key = self.flags & NegotiateFlags.REQUEST_NON_NT_SESSION_KEY)						
						self.setup_crypto()
						self.ntlmAuthenticate = NTLMAuthenticate.construct(self.flags, lm_response= self.ntlm_credentials.LMResponse, nt_response = self.ntlm_credentials.NTResponse, version = self.ntlmNegotiate.Version, encrypted_session = self.EncryptedRandomSessionKey)

							
							
				else:
					#NTLMv2
					# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/5e550938-91d4-459f-b67d-75d70009e3f3
					if self.settings.credential.is_guest == True:
						lmresp = LMResponse()
						lmresp.Response = b'\x00'
						self.ntlmAuthenticate = NTLMAuthenticate.construct(self.flags, lm_response= lmresp)
						return self.ntlmAuthenticate.to_bytes(), False
						
					else:
						#comment this out for testing!
						ti = self.ntlmChallenge.TargetInfo
						ti[AVPAIRType.MsvAvTargetName] = 'cifs/%s' % ti[AVPAIRType.MsvAvNbComputerName]
						###
						
						self.ntlm_credentials = netntlmv2.construct(self.ntlmChallenge.ServerChallenge, self.challenge, ti, self.settings.credential, timestamp = self.timestamp)
						self.KeyExchangeKey = self.ntlm_credentials.calc_key_exchange_key()						
						self.setup_crypto()
						
						#TODO: if "ti" / targetinfo in the challenge message has "MsvAvFlags" type and the bit for MIC is set (0x00000002) we need to send a MIC. probably...
						mic = None
						
						self.ntlmAuthenticate = NTLMAuthenticate.construct(self.flags, domainname= self.settings.credential.domain, workstationname= self.settings.credential.workstation, username= self.settings.credential.username, lm_response= self.ntlm_credentials.LMResponse, nt_response= self.ntlm_credentials.NTResponse, version = self.ntlmNegotiate.Version, encrypted_session = self.EncryptedRandomSessionKey, mic = mic)
						
				self.ntlmAuthenticate_raw = self.ntlmAuthenticate.to_bytes()
				return self.ntlmAuthenticate_raw, False
				
		elif self.mode.upper() == 'RELAY':
			if self.iteration_cnt == 0:
				self.ntlmNegotiate_raw = authData
				self.ntlmNegotiate = NTLMNegotiate.from_bytes(authData)	
				self.iteration_cnt += 1
			
			elif self.iteration_cnt == 1:
				self.ntlmChallenge_raw = authData
				self.ntlmChallenge = NTLMChallenge.from_bytes(authData)
				self.iteration_cnt += 1
			
			elif self.iteration_cnt == 2:
				self.ntlmChallenge_raw = authData
				self.ntlmChallenge = NTLMChallenge.from_bytes(authData)
				self.iteration_cnt += 1
				
			else:
				raise Exception('Too many iterations for relay mode!')
				


#def test_msdn():
#	credential = Credential()
#	credential.username = 'User'
#	credential.domain = 'Domain'
#	credential.password = 'Password'
#	
#	template = {
#			'flags'            :  NegotiateFlags.NEGOTIATE_56|
#								  NegotiateFlags.REQUEST_NON_NT_SESSION_KEY|
#								  NegotiateFlags.NEGOTIATE_KEY_EXCH|
#								  NegotiateFlags.NEGOTIATE_128|
#								  NegotiateFlags.NEGOTIATE_VERSION|
#								  NegotiateFlags.TARGET_TYPE_SERVER|
#								  NegotiateFlags.NEGOTIATE_ALWAYS_SIGN|
#								  NegotiateFlags.NEGOTIATE_NTLM|
#								  NegotiateFlags.NEGOTIATE_SIGN|
#								  NegotiateFlags.NEGOTIATE_SEAL|
#								  NegotiateFlags.NTLM_NEGOTIATE_OEM|
#								  NegotiateFlags.NEGOTIATE_UNICODE,
#			'version'          : Version.construct(WindowsMajorVersion.WINDOWS_MAJOR_VERSION_10, minor = WindowsMinorVersion.WINDOWS_MINOR_VERSION_0, build = 15063 ),
#			'domain_name'      : 'Domain',
#			'workstation_name' : 'COMPUTER',
#			'ntlm_downgrade'   : True,
#			'extended_security': False
#	}
#	settings = NTLMHandlerSettings(credential, mode = 'CLIENT', template_name = None, ntlm_downgrade = True, extended_security = False, custom_template = template)
#	handler = NTLMAUTHHandler(settings)
#	#assert handler.flags == int.from_bytes(b'\x33\x82\x02\xe2', "little", signed = False)
#	data, is_res = handler.authenticate(None)
#	print(data)
#	print(is_res)
#	
#	details = AVPairs({AVPAIRType.MsvAvNbDomainName: 'TEST', AVPAIRType.MsvAvNbComputerName: 'WIN2019AD', AVPAIRType.MsvAvDnsDomainName: 'test.corp', AVPAIRType.MsvAvDnsComputerName: 'WIN2019AD.test.corp', AVPAIRType.MsvAvTimestamp: b'\xae\xc6\x00\xbf\xc5\xfd\xd4\x01', AVPAIRType.MsvAvFlags: b'\x02\x00\x00\x00', AVPAIRType.MsvAvSingleHost: b"0\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00 \x00\x00R}'\xf24\xdet7`\x96c\x84\xd3oa\xae*\xa4\xfc*8\x06\x99\xf8\xca\xa6\x00\x01\x1bHm\x89", AVPAIRType.MsvChannelBindings: b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00', AVPAIRType.MsvAvTargetName: 'cifs/10.10.10.2'})
#	
#	challenge = NTLMChallenge.construct(challenge=b'\x01\x23\x45\x67\x89\xab\xcd\xef', targetName = 'Domain', targetInfo = details, version = handler.ntlmNegotiate.Version, flags= handler.flags)
#	data, is_res = handler.authenticate(challenge.to_bytes())
#	print(data)
#	print(is_res)
#	
#	print(handler.ntlmAuthenticate.LMChallenge.to_bytes().hex())
#	print(handler.ntlmAuthenticate.NTChallenge.to_bytes().hex())
#	
#
#def test():
#	template_name = 'Windows10_15063'
#	credential = Credential()
#	credential.username = 'test'
#	credential.password = 'test'
#	
#	settings = NTLMHandlerSettings(credential, mode = 'CLIENT', template_name = template_name, ntlm_downgrade = False, extended_security = True)
#	handler = NTLMAUTHHandler(settings)
#	data, is_res = handler.authenticate(None)
#	print(data)
#	print(is_res)
#	
#if __name__ == '__main__':
#	from aiosmb.ntlm.structures.version import Version, WindowsMajorVersion, WindowsMinorVersion
#	test_msdn()


import enum
import io

from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb.protocol.smb2.headers import *
from aiosmb.protocol.smb2.commands import *
from aiosmb.protocol.smb2.command_codes import *

class SMB2Compression:
	def __init__(self, header = None, data = None):
		self.header = header
		self.data   = data
	
	@staticmethod
	def from_bytes(bbuff):
		return SMB2Compression.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SMB2Compression()
		pos = buff.tell()
		t = buff.read(1)
		buff.seek(pos,0)
		if t == b'\xFC':
			msg.header = SMB2Header_COMPRESSION_TRANSFORM.from_buffer(buff)
		else:
			raise Exception('Unknown packet type for SMB2Compression! %s' % t)
		
		msg.data = buff.read()
		return msg

	def to_bytes(self):
		t = self.header.to_bytes()
		t += self.data
		return t

class SMB2Transform:
	def __init__(self, header = None, data = None):
		self.header = header
		self.data   = data
	
	@staticmethod
	def from_bytes(bbuff):
		return SMB2Transform.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SMB2Transform()
		pos = buff.tell()
		t = buff.read(1)
		buff.seek(pos,0)
		if t == b'\xFD':
			msg.header = SMB2Header_TRANSFORM.from_buffer(buff)
		else:
			raise Exception('Unknown packet type for SMB2Transform! %s' % t)
		
		msg.data = buff.read()
		return msg

	def to_bytes(self):
		t = self.header.to_bytes()
		t += self.data
		return t

class SMB2Message:
	def __init__(self,header = None,command = None ):
		self.header    = header
		self.command   = command

	@staticmethod
	def from_bytes(bbuff):
		return SMB2Message.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = SMB2Message()
		if SMB2Message.isAsync(buff):
			msg.header = SMB2Header_ASYNC.from_buffer(buff)
		else:
			msg.header = SMB2Header_SYNC.from_buffer(buff)
			
		# maybe it's an error...
		# not sure this is the best way to check fot he error message
		if msg.header.Status != NTStatus.SUCCESS and SMB2HeaderFlag.SMB2_FLAGS_SERVER_TO_REDIR in msg.header.Flags:
			if not (msg.header.Status == NTStatus.MORE_PROCESSING_REQUIRED and msg.header.Command.name == 'SESSION_SETUP'):
				pos = buff.tell()
				structure_size = int.from_bytes(buff.read(2), byteorder='little')
				buff.seek(pos, 0)
				if structure_size == 9:
					msg.command = ERROR_REPLY.from_buffer(buff)
					return msg

		classname = msg.header.Command.name
		try:
			if SMB2HeaderFlag.SMB2_FLAGS_SERVER_TO_REDIR in msg.header.Flags:
				classname += '_REPLY'
			else:
				classname += '_REQ'
			msg.command = command2object[classname].from_buffer(buff)
		except Exception as e:

			import traceback
			traceback.print_exc()
			#print('Could not find command implementation! %s' % str(e))
			msg.command = SMB2NotImplementedCommand.from_buffer(buff)

		return msg

	@staticmethod
	def isAsync(buff):
		"""
		jumping to the header flags and check if the AYSNC command flag is set
		"""
		pos = buff.tell()
		buff.seek(16, io.SEEK_SET)
		flags = SMB2HeaderFlag(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		buff.seek(pos, io.SEEK_SET)
		return SMB2HeaderFlag.SMB2_FLAGS_ASYNC_COMMAND in flags

	def to_bytes(self):
		t  = self.header.to_bytes()
		t += self.command.to_bytes()
		return t

	def __repr__(self):
		t = "== SMBv2 Message =="
		t += repr(self.header)
		t += repr(self.command)
		return t
		

command2object = {
	'NEGOTIATE_REQ'       : NEGOTIATE_REQ,
	'NEGOTIATE_REPLY'     : NEGOTIATE_REPLY,
	'SESSION_SETUP_REQ'     : SESSION_SETUP_REQ,
	'SESSION_SETUP_REPLY'     : SESSION_SETUP_REPLY,
	'TREE_CONNECT_REQ'     : TREE_CONNECT_REQ,
	'TREE_CONNECT_REPLY'     : TREE_CONNECT_REPLY,
	'CREATE_REQ'     : CREATE_REQ,
	'CREATE_REPLY'     : CREATE_REPLY,
	'READ_REQ'     : READ_REQ,
	'READ_REPLY'     : READ_REPLY,
	'QUERY_INFO_REPLY' : QUERY_INFO_REPLY,
	'QUERY_INFO_REQ' : QUERY_INFO_REQ,
	'QUERY_DIRECTORY_REQ' : QUERY_DIRECTORY_REQ,
	'QUERY_DIRECTORY_REPLY' : QUERY_DIRECTORY_REPLY,
	'TREE_DISCONNECT_REPLY' : TREE_DISCONNECT_REPLY,
	'TREE_DISCONNECT_REQ' : TREE_DISCONNECT_REQ,
	'ECHO_REQ' : ECHO_REQ,
	'ECHO_REPLY' : ECHO_REPLY,
	'LOGOFF_REQ'   : LOGOFF_REQ,
	'LOGOFF_REPLY' : LOGOFF_REPLY,
	'ERROR_REPLY' : ERROR_REPLY,
	'CLOSE_REPLY' : CLOSE_REPLY,
	'WRITE_REPLY' : WRITE_REPLY,
	'WRITE_REQ' : WRITE_REQ,
	'FLUSH_REPLY' : FLUSH_REPLY,
	'FLUSH_REQ' : FLUSH_REQ,
}


from pathlib import PureWindowsPath
from aiosmb.wintypes.access_mask import *
from aiosmb.protocol.smb2.commands import *
from aiosmb.wintypes.fscc.structures.fileinfoclass import FileInfoClass
from aiosmb.protocol.smb2.commands.query_info import SecurityInfo
import io


class SMBFile:
	def __init__(self):
		self.tree_id = None
		self.parent_dir = None
		self.fullpath = None
		self.unc_path = None
		self.share_path = None
		self.name = None
		self.size = None
		self.creation_time = None
		self.last_access_time = None
		self.last_write_time = None
		self.change_time = None
		self.allocation_size = None
		self.attributes = None
		self.file_id = None
		self.sid = None

		#internal
		self.__connection = None
		self.__position = 0
		self.is_pipe = False

	@staticmethod
	def from_uncpath(unc_path):
		"""
		Creates SMBFile object from the UNC path supplied.
		Example uncpath: \\\\127.0.0.1\\C$\\temp\\test.exe
		"""
		unc = PureWindowsPath(unc_path)
		f = SMBFile()
		f.share_path = unc.drive
		f.fullpath = '\\'.join(unc.parts[1:])
		
		return f

	@staticmethod
	def from_remotepath(connection, remotepath):
		"""
		Creates SMBFile object from the connection and the remote path supplied.
		Example remotepath: \\C$\\temp\\test.exe
		"""
		temp = '\\\\%s\\%s'
		if remotepath[0] == '\\':
			temp = '\\\\%s%s'
		unc = temp % (connection.target.get_hostname_or_ip(), remotepath)
		return SMBFile.from_uncpath(unc)

	@staticmethod
	async def delete(connection, remotepath):
		remfile = SMBFile.from_remotepath(connection, remotepath)
		tree_entry = await connection.tree_connect(remfile.share_path)
		tree_id = tree_entry.tree_id

		desired_access = FileAccessMask.DELETE | FileAccessMask.FILE_READ_ATTRIBUTES
		share_mode = ShareAccess.FILE_SHARE_DELETE
		create_options = CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_DELETE_ON_CLOSE 
		create_disposition = CreateDisposition.FILE_OPEN
		file_attrs = 0

		file_id = await connection.create(tree_id, remfile.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs, return_reply = False)
		if file_id is not None:
			await connection.close(tree_id, file_id)

		await connection.tree_disconnect(tree_id)
		return True

	async def get_security_descriptor(self, connection):
		if self.sid is None:
			file_id = None
			try:
				desired_access = FileAccessMask.READ_CONTROL
				share_mode = ShareAccess.FILE_SHARE_READ
				create_options = CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_SYNCHRONOUS_IO_NONALERT 
				file_attrs = 0
				create_disposition = CreateDisposition.FILE_OPEN
				file_id = await connection.create(self.tree_id, self.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs)
				
				self.sid = await connection.query_info(
					self.tree_id,
					file_id,
					info_type = QueryInfoType.SECURITY, 
					information_class = FileInfoClass.NONE, 
					additional_information = SecurityInfo.ATTRIBUTE_SECURITY_INFORMATION | SecurityInfo.DACL_SECURITY_INFORMATION | SecurityInfo.OWNER_SECURITY_INFORMATION | SecurityInfo.GROUP_SECURITY_INFORMATION, 
					flags = 0, 
				)
			except:
				raise

			finally:
				if file_id is not None:
					await connection.close(self.tree_id, file_id)

			return self.sid

	async def __read(self, size, offset):
		"""
		This is the main function for reading.
		It does not do buffering, so if more data is returned it will just discard it
		If less data is returned than requested it will do more reads until the requested size is reached.
		Do not call this directly as it could go in an infinite loop 
		"""
		if self.is_pipe == True:
			data, remaining = await self.__connection.read(self.tree_id, self.file_id, offset = offset, length = size)
			return data
		
		buffer = b''
		if size > self.__connection.MaxReadSize:
			i, rem = divmod(size, self.__connection.MaxReadSize)
			for _ in range(i+1):
				data, remaining = await self.__connection.read(self.tree_id, self.file_id, offset = offset, length = self.__connection.MaxReadSize)
				offset += len(data)
				buffer += data
			
			return buffer[:size]
		else:
			data, remaining = await self.__connection.read(self.tree_id, self.file_id, offset = offset, length = self.__connection.MaxReadSize)
			buffer += data
			
			return buffer[:size]

	async def __write(self, data, position_in_file = 0):
		"""
		Data must be bytes
		"""
		remaining = len(data)
		total_bytes_written = 0
		offset = 0
		
		while remaining != 0:
			bytes_written = await self.__connection.write(self.tree_id, self.file_id, data[offset:len(data)], offset = position_in_file + offset)
			total_bytes_written += bytes_written
			remaining -= bytes_written
			offset += bytes_written
		
		return total_bytes_written

	async def open(self, connection, mode = 'r'):
		self.__connection = connection
		self.mode = mode
		if 'p' in self.mode:
			self.is_pipe = True
			self.size = 0
		
		if not self.tree_id:
			tree_entry = await connection.tree_connect(self.share_path)
			self.tree_id = tree_entry.tree_id
		

		#then connect to file
		if 'r' in mode and 'w' in mode:
			raise ValueError('must have exactly one of read/write mode')
			
		if 'r' in mode:
			desired_access = FileAccessMask.FILE_READ_DATA | FileAccessMask.FILE_READ_ATTRIBUTES
			share_mode = ShareAccess.FILE_SHARE_READ
			create_options = CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_SYNCHRONOUS_IO_NONALERT 
			file_attrs = 0
			create_disposition = CreateDisposition.FILE_OPEN
			
			self.file_id, smb_reply = await connection.create(self.tree_id, self.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs, return_reply = True)
			self.size = smb_reply.EndofFile
			
		elif 'w' in mode:
			desired_access = FileAccessMask.GENERIC_READ | FileAccessMask.GENERIC_WRITE
			share_mode = ShareAccess.FILE_SHARE_READ | ShareAccess.FILE_SHARE_WRITE
			create_options = CreateOptions.FILE_NON_DIRECTORY_FILE | CreateOptions.FILE_SYNCHRONOUS_IO_NONALERT 
			file_attrs = 0
			create_disposition = CreateDisposition.FILE_OPEN_IF #FILE_OPEN ? might cause an issue?
			
			self.file_id, smb_reply = await connection.create(self.tree_id, self.fullpath, desired_access, share_mode, create_options, create_disposition, file_attrs, return_reply = True)
			self.size = smb_reply.EndofFile
			
		else:
			raise Exception('ONLY read and write is supported at the moment!')
		
		
	async def seek(self, offset, whence = 0):
		if whence == 0:
			if offset < 0:
				raise Exception('Offset must be > 0 when whence is 0')
			if offset > self.size:
				raise Exception('Seeking outside of file size!')
			self.__position = offset
		
		elif whence == 1:
			if 0 < self.__position + offset < self.size:
				self.__position += offset
			else:
				raise Exception('Seeking outside of file size!')
		
		elif whence == 2:
			if 0 < self.size + offset < self.size:
				self.__position = self.size + offset
			else:
				raise Exception('Seeking outside of file size!')
		
	async def read(self, size = -1):
		if self.is_pipe is True:
			data = await self.__read(size, 0)
			return data
		
		if size > self.size:
			raise Exception('Requested read size %s is larger than the file size %s' % (hex(size), hex(self.size)))

		if size == 0:
			raise Exception('Cant read 0 bytes')
			
		elif size == -1:
			data = await self.__read(self.size - self.__position, self.__position)
			self.__position += len(data)
			
			return data
			
		elif size > 0:
			if self.__position == self.size:
				return None
			if size + self.__position > self.size:
				size = self.size - self.__position
			data = await self.__read(size, self.__position)
			self.__position += len(data)
			return data

	async def read_chunked(self, size = -1, chunksize = -1):
		"""
		Much like read, but yields chauks of chunksize untill the full size is read
		Use this when reading large files as a whole, as this method doesn't caches 
		the whole data read in memory, only chunskize
		chunksize -1 will use the maximum chunk allowed by the underlying connection layer, it is advised to not set manually!
		"""
		if self.is_pipe is True:
			raise Exception('Cant stream pipes!')

		if chunksize == -1:
			chunksize = self.__connection.MaxReadSize

		if size == 0:
			yield None
			
		elif size == -1:
			while True:
				req_size = chunksize
				if self.size - self.__position < chunksize:
					req_size = self.size - self.__position
					if req_size == 0:
						#consumed all data
						yield None
						raise StopIteration

				data = await self.__read(req_size, self.__position)
				self.__position += len(data)
				yield data
			
		elif size > 0:
			if self.__position == self.size:
				yield None
				raise StopIteration
			if size + self.__position > self.size:
				size = self.size - self.__position

			while True:
				req_size = chunksize
				if size - self.__position < chunksize:
					req_size = size - self.__position
				if req_size == 0:
					yield None
					raise StopIteration
				data = await self.__read(req_size, self.__position)
				self.__position += len(data)
				yield data
			
	async def write(self, data):
		if len(data) < self.__connection.MaxWriteSize:
			count = await self.__write(data, self.__position)
			if self.is_pipe == False:
				self.__position += count
			return count

		total_writen = 0

		while total_writen != len(data) :
			count = await self.__write(data[total_writen:total_writen+self.__connection.MaxWriteSize], self.__position)
			total_writen += count
			if self.is_pipe == False:
				self.__position += count		

		return total_writen

	async def write_buffer(self, buffer):
		"""
		Doesnt work with pipes!
		"""
		if self.is_pipe == True:
			raise Exception('Doesnt work with pipes!')
		
		total_writen = 0
		while True:
			chunk = buffer.read(self.__connection.MaxWriteSize)
			if len(chunk) == 0:
				return total_writen
			bytes_written = await self.__write(chunk, self.__position)

			self.__position += bytes_written
			total_writen += bytes_written
		
	async def flush(self):
		if 'r' in self.mode:
			return
		else:
			await self.__connection.flush(self.tree_id, self.file_id)
		
	async def close(self):
		await self.flush()
		await self.__connection.close(self.tree_id, self.file_id)
	
		
	def __str__(self):
		t = '===== FILE =====\r\n'
		for k in self.__dict__:
			if k.startswith('parent_'):
				continue
			t += '%s : %s\r\n' % (k, self.__dict__[k])
		
		return t


#
#
# This is just a simple interface to the minikerberos library to support SPNEGO
# 
#
# - Hardships - 
# 1. DCERPC kerberos authentication requires a complete different approach and flags,
#    also requires mutual authentication
#
# - Links - 
# 1. Most of the idea was taken from impacket
# 2. See minikerberos library

import datetime

from minikerberos.common import *

from minikerberos.protocol.asn1_structs import AP_REP, EncAPRepPart, EncryptedData
from minikerberos.gssapi.gssapi import get_gssapi
from minikerberos.protocol.structures import ChecksumFlags
from minikerberos.protocol.encryption import Enctype, Key, _enctype_table
from minikerberos.protocol.constants import MESSAGE_TYPE
from minikerberos.aioclient import AIOKerberosClient

# SMBKerberosCredential

class SMBKerberos:
	def __init__(self, settings):
		self.settings = settings
		self.mode = None
		self.ccred = None
		self.target = None
		self.spn = None
		self.kc = None
		
		self.session_key = None
		self.gssapi = None
		self.iterations = 0
		self.etype = None
	
		self.setup()
		
		
	async def sign(self, data, message_no, direction = 'init'):
		return self.gssapi.GSS_GetMIC(data, message_no, direction = direction)	
		
	async def encrypt(self, data, message_no):
		return self.gssapi.GSS_Wrap(data, message_no)
		
	async def decrypt(self, data, message_no, direction='init', auth_data=None):
		return self.gssapi.GSS_Unwrap(data, message_no, direction=direction, auth_data=auth_data)
		
	def setup(self):
		self.mode = self.settings.mode
		self.ccred = self.settings.ccred
		self.spn = self.settings.spn
		self.target = self.settings.target
		
		self.kc = AIOKerberosClient(self.ccred, self.target)
		
	
	def get_session_key(self):
		return self.session_key.contents
	
	async def authenticate(self, authData, flags = None, seq_number = 0, is_rpc = False):

		if self.iterations == 0:
			#tgt = await self.kc.get_TGT(override_etype=[18])
			tgt = await self.kc.get_TGT()
			tgs, encpart, self.session_key = await self.kc.get_TGS(self.spn)
		ap_opts = []
		if is_rpc == True:
			if self.iterations == 0:
				ap_opts.append('mutual-required')
				flags = ChecksumFlags.GSS_C_CONF_FLAG | ChecksumFlags.GSS_C_INTEG_FLAG | ChecksumFlags.GSS_C_SEQUENCE_FLAG|\
						ChecksumFlags.GSS_C_REPLAY_FLAG | ChecksumFlags.GSS_C_MUTUAL_FLAG | ChecksumFlags.GSS_C_DCE_STYLE
						
				apreq = self.kc.construct_apreq(tgs, encpart, self.session_key, flags = flags, seq_number = seq_number, ap_opts=ap_opts)					
				self.iterations += 1
				return apreq, False
				
			else:
				#mutual authentication part here
				aprep = AP_REP.load(authData).native
				cipher = _enctype_table[int(aprep['enc-part']['etype'])]()
				cipher_text = aprep['enc-part']['cipher']
				temp = cipher.decrypt(self.session_key, 12, cipher_text)
				
				enc_part = EncAPRepPart.load(temp).native
				cipher = _enctype_table[int(enc_part['subkey']['keytype'])]()
				
				now = datetime.datetime.now(datetime.timezone.utc)
				apreppart_data = {}
				apreppart_data['cusec'] = now.microsecond
				apreppart_data['ctime'] = now.replace(microsecond=0)
				apreppart_data['seq-number'] = enc_part['seq-number']
				
				apreppart_data_enc = cipher.encrypt(self.session_key, 12, EncAPRepPart(apreppart_data).dump(), None)
				
				#overriding current session key
				self.session_key = Key(cipher.enctype, enc_part['subkey']['keyvalue'])
				
				ap_rep = {}
				ap_rep['pvno'] = 5 
				ap_rep['msg-type'] = MESSAGE_TYPE.KRB_AP_REP.value
				ap_rep['enc-part'] = EncryptedData({'etype': self.session_key.enctype, 'cipher': apreppart_data_enc}) 
				
				token = AP_REP(ap_rep).dump()
				self.gssapi = get_gssapi(self.session_key)
				self.iterations += 1
				
				return token, False
		else:
			apreq = self.kc.construct_apreq(tgs, encpart, self.session_key, flags = flags, seq_number = seq_number, ap_opts=ap_opts)
			return apreq, False


import enum
import copy

from aiosmb.dcerpc.v5.common.connection.connectionstring import DCERPCStringBinding


class DCERPCTargetType:
	UDP = 'UDP'
	TCP = 'TCP'
	SMB = 'SMB'
	HTTP = 'HTTP'
	LOCAL = 'LOCAL'


class DCERPCTarget:
	def __init__(self, connection_string, ttype, proxy = None, timeout = 1):
		self.ip = None
		self.connection_string = connection_string
		self.type = ttype
		self.timeout = timeout
		self.proxy = None

	def get_hostname_or_ip(self):
		return self.ip

	@staticmethod
	def from_connection_string(s, smb_connection = None, timeout = 1):
		if isinstance(s, str):
			connection_string = DCERPCStringBinding(s)
		elif isinstance(s, DCERPCStringBinding):
			connection_string = s
		else:
			raise Exception('Unknown string binding type %s' % type(s))
			
		na = connection_string.get_network_address()
		ps = connection_string.get_protocol_sequence()
		if ps == 'ncadg_ip_udp':
			raise Exception('DCERPC UDP not implemented')
			port = connection_string.get_endpoint()
			target = DCERPCUDPTarget(connection_string, na, int(port), timeout = timeout)
		elif ps == 'ncacn_ip_tcp':
			port = connection_string.get_endpoint()
			target = DCERPCTCPTarget(connection_string, na, port, timeout = timeout)
		elif ps == 'ncacn_http':
			raise Exception('DCERPC HTTP not implemented')
			target = DCERPCHTTPTarget(connection_string, na, int(port), timeout = timeout)
		elif ps == 'ncacn_np':
			named_pipe = connection_string.get_endpoint()
			if named_pipe:
				named_pipe = named_pipe[len(r'\pipe'):]
				target = DCERPCSMBTarget(connection_string, na, named_pipe, smb_connection=smb_connection, timeout = timeout)
			else:
				 target = DCERPCSMBTarget(connection_string, na, smb_connection=smb_connection, timeout = timeout)
		elif ps == 'ncalocal':
			raise Exception('DCERPC LOCAL not implemented')
			target = DCERPCLocalTarget(connection_string, na, int(port), timeout = timeout)
		
		else:
			raise Exception('Unknown DCERPC protocol %s' % ps)

		if smb_connection.target.proxy is not None:
			target.proxy = copy.deepcopy(smb_connection.target.proxy)
			
		return target

	def __str__(self):
		t = '==== DCERPCTarget ====\r\n'
		for k in self.__dict__:
			t += '%s: %s\r\n' % (k, self.__dict__[k])
			
		return t


class DCERPCTCPTarget(DCERPCTarget):
	def __init__(self, connection_string, ip, port, timeout = 1):
		DCERPCTarget.__init__(self, connection_string, DCERPCTargetType.TCP, timeout = timeout)
		self.ip = ip
		self.port = int(port)

class DCERPCUDPTarget(DCERPCTarget):
	def __init__(self, connection_string, ip, port, timeout = 1):
		DCERPCTarget.__init__(self, connection_string, DCERPCTargetType.UDP, timeout = timeout)
		self.ip = ip
		self.port = int(port)

class DCERPCSMBTarget(DCERPCTarget):
	def __init__(self, connection_string, ip, pipe = None, smb_connection = None, timeout = 1):
		DCERPCTarget.__init__(self, connection_string, DCERPCTargetType.SMB, timeout = timeout)
		self.ip = ip
		self.pipe = pipe
		self.smb_connection = smb_connection #storing the smb connection is already exists...

class DCERPCHTTPTarget(DCERPCTarget):
	def __init__(self, connection_string, ip, port, timeout = 1):
		DCERPCTarget.__init__(self, connection_string, DCERPCTargetType.HTTP, timeout = timeout)
		self.ip = ip
		self.port = int(port)

class DCERPCLocalTarget(DCERPCTarget):
	def __init__(self, connection_string, ip, port, timeout = 1):
		DCERPCTarget.__init__(self, connection_string, DCERPCTargetType.LOCAL, timeout = timeout)
		self.ip = ip
		self.port = int(port)


if __name__ == '__main__':
	s = ''
	target = DCERPCTarget.from_connection_string(s)
	





# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Description:
#   Generate UUID compliant with http://www.webdav.org/specs/draft-leach-uuids-guids-01.txt.
#   A different, much simpler (not necessarily better) algorithm is used.
#
# Author:
#   Javier Kohen (jkohen)
#
from __future__ import absolute_import
from __future__ import print_function
import re

from random import randrange
from struct import pack, unpack

def generate():
    # UHm... crappy Python has an maximum integer of 2**31-1.
    top = (1<<31)-1
    return pack("IIII", randrange(top), randrange(top), randrange(top), randrange(top))

def bin_to_string(uuid):
    uuid1, uuid2, uuid3 = unpack('<LHH', uuid[:8])
    uuid4, uuid5, uuid6 = unpack('>HHL', uuid[8:16])
    return '%08X-%04X-%04X-%04X-%04X%08X' % (uuid1, uuid2, uuid3, uuid4, uuid5, uuid6)

def string_to_bin(uuid):
    matches = re.match('([\dA-Fa-f]{8})-([\dA-Fa-f]{4})-([\dA-Fa-f]{4})-([\dA-Fa-f]{4})-([\dA-Fa-f]{4})([\dA-Fa-f]{8})', uuid)
    (uuid1, uuid2, uuid3, uuid4, uuid5, uuid6) = [int(x, 16) for x in matches.groups()]
    uuid = pack('<LHH', uuid1, uuid2, uuid3)
    uuid += pack('>HHL', uuid4, uuid5, uuid6)
    return uuid

def stringver_to_bin(s):
    (maj,min) = s.split('.')
    return pack('<H',int(maj)) + pack('<H',int(min))

def uuidtup_to_bin(tup):
    if len(tup) != 2:
        return
    return string_to_bin(tup[0]) + stringver_to_bin(tup[1])

def bin_to_uuidtup(bin):
    assert len(bin) == 20
    uuidstr = bin_to_string(bin[:16])
    maj, min = unpack("<HH", bin[16:])
    return uuidstr, "%d.%d" % (maj, min)

#input: string
#output: tuple (uuid,version) 
#if version is not found in the input string "1.0"  is returned
#example: 
#           "00000000-0000-0000-0000-000000000000 3.0" returns ('00000000-0000-0000-0000-000000000000','3.0') 
#           "10000000-2000-3000-4000-500000000000 version 3.0" returns ('00000000-0000-0000-0000-000000000000','3.0') 
#           "10000000-2000-3000-4000-500000000000 v 3.0" returns ('00000000-0000-0000-0000-000000000000','3.0') 
#           "10000000-2000-3000-4000-500000000000" returns ('00000000-0000-0000-0000-000000000000','1.0') 
def string_to_uuidtup(s):
    g =  re.search("([A-Fa-f0-9]{8}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{4}-[A-Fa-f0-9]{12}).*?([0-9]{1,5}\.[0-9]{1,5})",s+" 1.0")
    if g: 
        (u,v) = g.groups()
        return (u,v)
    return

def uuidtup_to_string(tup):
    uuid, (maj, min) = tup
    return "%s v%d.%d" % (uuid, maj, min)



"""
The idea here is to offer compatibility with 3rd party libraries by extending wrappers for ech encryption mode
This is needed because the pure python implementation for encryption and hashing algorithms are quite slow

currently it's not the perfect wrapper, needs to be extended
"""

from aiosmb.crypto.BASE import symmetricBASE, cipherMODE, padMode
import aiosmb.crypto.pure.DES.DES as _pyDES
try:
	from Crypto.Cipher import DES3 as _pyCryptoDES3
except:
	pass

try:
	from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
	from cryptography.hazmat.primitives import padding
	from cryptography.hazmat.backends import default_backend
except:
	pass

class pureTDES(symmetricBASE):
	def __init__(self, key, mode = cipherMODE.ECB, IV = None, pad = None, padMode = None):
		symmetricBASE.__init__(self)
		if not isinstance(key, bytes):
			raise Exception('Key needs to be bytes!')

		self.mode = mode
		self.IV = IV
		self.pad = pad
		self.padMode = padMode

	def setup_cipher(self):
		if self.mode == cipherMODE.ECB:
			mode = _pyDes.ECB
			self._cipher = _pyDES.triple_des(self.key, mode)
		elif self.mode == cipherMODE.CBC:
			mode = _pyDES.CBC
			if padMode is None:
				self._cipher = _pyDES.triple_des(self.key, mode, self.IV, self.pad, self.padmode)
			else:
				self._cipher = _pyDES.triple_des(self.key, mode, self.IV)
		else:
			raise Exception('Unknown cipher mode!')

	def encrypt(self, data):
		return self._cipher.encrypt(data)
	def decrypt(self, data):
		return self._cipher.decrypt(data)

class pyCryptoTDES(symmetricBASE):
	def __init__(self, key, mode = cipherMODE.ECB, IV = None, pad = None, padMode = None):
		self.key = key
		self.mode = mode
		self.IV = IV
		self.pad = pad
		self.padMode = padMode

		symmetricBASE.__init__(self)
		
	def setup_cipher(self):
		if self.mode == cipherMODE.ECB:
			self._cipher = _pyCryptoDES3.new(self.key, _pyCryptoDES3.MODE_ECB)
		
		elif self.mode == cipherMODE.CBC:
			self._cipher = _pyCryptoDES3.new(self.key, _pyCryptoDES3.MODE_CBC, self.IV)
		else:
			raise Exception('Unknown cipher mode!')
		
	def encrypt(self, data):
		return self._cipher.encrypt(data)
	def decrypt(self, data):
		return self._cipher.decrypt(data)

class cryptographyTDES(symmetricBASE):
	def __init__(self, key, mode = cipherMODE.ECB, IV = None, pad = None, padMode = None):
		if not isinstance(key, bytes):
			raise Exception('Key needs to be bytes!')
		self.IV = IV
		if mode == cipherMode.ECB:
			self.IV = modes.ECB()
		elif mode == cipherMODE.CBC:
			self.IV = modes.CBC(IV)
		else:
			raise Exception('Unknown cipher mode!')
		self.key = key
		
		""" TODO padding
		if self.padMode is not None:
		"""
		
		self.encryptor = None
		self.decryptor = None
		symmetricBASE.__init__(self)

	def setup_cipher(self):
		algorithm = algorithms.TripleDES(self.key)
		self._cipher = Cipher(algorithm, mode=self.IV, backend=default_backend())
		self.encryptor = self._cipher.encryptor()
		self.decryptor = self._cipher.decryptor()

	def encrypt(self, data):
		return self.encryptor.update(data)
	

	def decrypt(self, data):
		return self.decryptor.update(data)

